from assay import assert_raises
from numpy import abs, arange, sqrt

from skyfield import constants
from skyfield.api import Distance, load, wgs84, wms
from skyfield.functions import length_of
from skyfield.positionlib import Apparent, Barycentric
from skyfield.toposlib import ITRSPosition, iers2010

angle = (-15, 15, 35, 45)

def ts():
    yield load.timescale()

def t():
    ts = load.timescale()
    yield ts.utc(2020, 11, 3, 17, 5)
    yield ts.utc(2020, 11, 3, 17, [5, 5])

def test_polar_radius():
    assert round(wgs84.polar_radius.km) == 6357

def test_latitude_longitude_elevation_str_and_repr():
    w = wgs84.latlon(36.7138, -112.2169, 2400.0)
    assert str(w) == ('WGS84 latitude +36.7138 N'
                      ' longitude -112.2169 E elevation 2400.0 m')
    assert repr(w) == ('<GeographicPosition WGS84 latitude +36.7138 N'
                       ' longitude -112.2169 E elevation 2400.0 m>')

    w = wgs84.latlon([1.0, 2.0], [3.0, 4.0], [5.0, 6.0])
    assert str(w) == (
        'WGS84 latitude [+1.0000 +2.0000] N'
        ' longitude [3.0000 4.0000] E'
        ' elevation [5.0 6.0] m'
    )
    assert repr(w) == '<GeographicPosition {0}>'.format(w)

    w = wgs84.latlon(arange(6.0), arange(10.0, 16.0), arange(20.0, 26.0))
    assert str(w) == (
        'WGS84 latitude [+0.0000 +1.0000 ... +4.0000 +5.0000] N'
        ' longitude [10.0000 11.0000 ... 14.0000 15.0000] E'
        ' elevation [20.0 21.0 ... 24.0 25.0] m'
    )
    assert repr(w) == '<GeographicPosition {0}>'.format(w)

def test_raw_itrs_position():
    d = Distance(au=[1, 2, 3])
    p = ITRSPosition(d)
    ts = load.timescale()
    t = ts.utc(2020, 12, 16, 12, 59)
    p.at(t)

def test_wgs84_velocity_matches_actual_motion():
    # It looks like this is a sweet spot for accuracy: presumably a
    # short enough fraction of a second that the vector does not time to
    # change direction much, but long enough that the direction does not
    # get lost down in the noise.
    factor = 300.0

    ts = load.timescale()
    t = ts.utc(2019, 11, 2, 3, 53, [0, 1.0 / factor])
    jacob = wgs84.latlon(36.7138, -112.2169)
    p = jacob.at(t)
    velocity1 = p.position.km[:,1] - p.position.km[:,0]
    velocity2 = p.velocity.km_per_s[:,0]
    assert length_of(velocity2 - factor * velocity1) < 0.0007

def test_lst():
    ts = load.timescale()
    ts.delta_t_table = [-1e99, 1e99], [69.363285] * 2  # from finals2000A.all
    t = ts.utc(2020, 11, 27, 15, 34)
    top = wgs84.latlon(0.0, 0.0)
    expected = 20.0336663100  # see "authorities/horizons-lst"
    actual = top.lst_hours_at(t)
    difference_mas = (actual - expected) * 3600 * 15 * 1e3
    horizons_ra_offset_mas = 51.25
    difference_mas -= horizons_ra_offset_mas
    assert abs(difference_mas) < 1.0

def test_itrs_xyz_attribute_and_itrf_xyz_method():
    top = wgs84.latlon(45.0, 0.0, elevation_m=constants.AU_M - constants.ERAD)

    x, y, z = top.itrs_xyz.au
    assert abs(x - sqrt(0.5)) < 2e-7
    assert abs(y - 0.0) < 1e-14
    assert abs(z - sqrt(0.5)) < 2e-7

    ts = load.timescale()
    t = ts.utc(2019, 11, 2, 3, 53)
    x, y, z = top.at(t).itrf_xyz().au
    assert abs(x - sqrt(0.5)) < 1e-4
    assert abs(y - 0.0) < 1e-14
    assert abs(z - sqrt(0.5)) < 1e-4

def test_polar_motion_when_computing_topos_position(ts):
    xp_arcseconds = 11.0
    yp_arcseconds = 22.0
    ts.polar_motion_table = [0.0], [xp_arcseconds], [yp_arcseconds]

    top = iers2010.latlon(wms(42, 21, 24.1), wms(-71, 3, 24.8), 43.0)
    t = ts.utc(2005, 11, 12, 22, 2)

    # "expected" comes from:
    # from novas.compat import ter2cel
    # print(ter2cel(t.whole, t.ut1_fraction, t.delta_t, xp_arcseconds,
    #               yp_arcseconds, top.itrs_xyz.km, method=1))

    expected = (3129.530248036487, -3535.1665884086683, 4273.94957733827)
    assert max(abs(top.at(t).position.km - expected)) < 3e-11

def test_polar_motion_when_computing_altaz_coordinates(ts):
    latitude = 37.3414
    longitude = -121.6429
    elevation = 1283.0
    ra_hours = 5.59
    dec_degrees = -5.45

    xp_arcseconds = 11.0
    yp_arcseconds = 22.0
    ts.polar_motion_table = [0.0], [xp_arcseconds], [yp_arcseconds]

    t = ts.utc(2020, 11, 12, 22, 16)
    top = wgs84.latlon(latitude, longitude, elevation)

    pos = Apparent.from_radec(ra_hours, dec_degrees, epoch=t)
    pos.t = t
    pos.center = top

    alt, az, distance = pos.altaz()

    # To generate the test altitude and azimuth below:
    # from novas.compat import equ2hor, make_on_surface
    # location = make_on_surface(latitude, longitude, elevation, 0, 0)
    # (novas_zd, novas_az), (rar, decr) = equ2hor(
    #     t.ut1, t.delta_t, xp_arcseconds, yp_arcseconds, location,
    #     ra_hours, dec_degrees, 0,
    # )
    # novas_alt = 90.0 - novas_zd
    # print(novas_alt, novas_az)

    novas_alt = -58.091983295564205
    novas_az = 1.8872567543791035

    assert abs(alt.degrees - novas_alt) < 1.9e-9
    assert abs(az.degrees - novas_az) < 1.3e-7

def test_subpoint_with_wrong_center(ts, angle):
    t = ts.utc(2020, 12, 31)
    p = Barycentric([0,0,0], t=t)
    with assert_raises(ValueError, 'you can only calculate a geographic'
                       ' position from a position which is geocentric'
                       ' .center=399., but this position has a center of 0'):
        wgs84.subpoint(p)

def test_iers2010_subpoint(ts, angle):
    t = ts.utc(2018, 1, 19, 14, 37, 55)
    # An elevation of 0 is more difficult for the routine's accuracy
    # than a very large elevation.
    top = iers2010.latlon(angle, angle, elevation_m=0.0)
    p = top.at(t)
    b = iers2010.subpoint(p)

    error_degrees = abs(b.latitude.degrees - angle)
    error_mas = 60.0 * 60.0 * 1000.0 * error_degrees
    assert error_mas < 0.1

    error_degrees = abs(b.longitude.degrees - angle)
    error_mas = 60.0 * 60.0 * 1000.0 * error_degrees
    assert error_mas < 0.1

def test_wgs84_subpoint(ts, angle):
    t = ts.utc(2018, 1, 19, 14, 37, 55)
    # An elevation of 0 is more difficult for the routine's accuracy
    # than a very large elevation.
    top = wgs84.latlon(angle, angle, elevation_m=0.0)
    p = top.at(t)
    b = wgs84.subpoint(p)

    error_degrees = abs(b.latitude.degrees - angle)
    error_mas = 60.0 * 60.0 * 1000.0 * error_degrees
    assert error_mas < 0.1

    error_degrees = abs(b.longitude.degrees - angle)
    error_mas = 60.0 * 60.0 * 1000.0 * error_degrees
    assert error_mas < 0.1

def test_wgs84_subpoint_at_pole(ts):
    # The `height` previously suffered from very low precision at the pole.
    t = ts.utc(2023, 4, 7, 12, 44)
    p = wgs84.latlon(90, 0, elevation_m=10.0).at(t)
    micrometer = 1e-6

    h = wgs84.height_of(p)
    assert abs(h.m - 10.0) < micrometer

    g = wgs84.geographic_position_of(p)
    assert abs(g.elevation.m - 10.0) < micrometer

def test_wgs84_round_trip_with_polar_motion(ts, angle):
    t = ts.utc(2018, 1, 19, 14, 37, 55)
    ts.polar_motion_table = [0.0], [0.003483], [0.358609]

    top = wgs84.latlon(angle, angle, elevation_m=0.0)
    p = top.at(t)
    b = wgs84.subpoint(p)

    error_degrees = abs(b.latitude.degrees - angle)
    error_mas = 60.0 * 60.0 * 1000.0 * error_degrees
    assert error_mas < 0.1

    error_degrees = abs(b.longitude.degrees - angle)
    error_mas = 60.0 * 60.0 * 1000.0 * error_degrees
    assert error_mas < 0.1

def test_latlon_and_subpoint_methods(t, angle):
    g = wgs84.latlon(angle, 2 * angle, elevation_m=1234.0)
    pos = g.at(t)
    x = all if t.shape else lambda x: x

    def check_lat(lat): assert x(abs(g.latitude.mas() - lat.mas()) < 0.1)
    def check_lon(lon): assert x(abs(g.longitude.mas() - lon.mas()) < 0.1)
    def check_height(h): assert x(abs(g.elevation.m - h.m) < 1e-7)
    def check_itrs(xyz, expected_distance):
        r1 = g.itrs_xyz.m
        r2 = xyz
        if len(r2.shape) > len(r1.shape):
            r1.shape += (1,)
        actual_distance = length_of(r1 - r2)
        assert x(abs(actual_distance - expected_distance) < 1e-7)

    lat, lon = wgs84.latlon_of(pos)
    check_lat(lat)
    check_lon(lon)

    height = wgs84.height_of(pos)
    check_height(height)

    g2 = wgs84.geographic_position_of(pos)
    check_lat(g2.latitude)
    check_lon(g2.longitude)
    check_height(g2.elevation)
    check_itrs(g2.itrs_xyz.m, 0.0)

    g2 = wgs84.subpoint(pos)  # old deprecated method name
    check_lat(g2.latitude)
    check_lon(g2.longitude)
    check_height(g2.elevation)
    check_itrs(g2.itrs_xyz.m, 0.0)

    g2 = wgs84.subpoint_of(pos)
    check_lat(g2.latitude)
    check_lon(g2.longitude)
    assert g2.elevation.m == 0.0
    check_itrs(g2.itrs_xyz.m, 1234.0)

def test_deprecated_position_subpoint_method(ts, angle):
    t = ts.utc(2018, 1, 19, 14, 37, 55)
    top = iers2010.latlon(angle, angle, elevation_m=0.0)
    b = top.at(t).subpoint()

    error_degrees = abs(b.latitude.degrees - angle)
    error_mas = 60.0 * 60.0 * 1000.0 * error_degrees
    assert error_mas < 0.1

    error_degrees = abs(b.longitude.degrees - angle)
    error_mas = 60.0 * 60.0 * 1000.0 * error_degrees
    assert error_mas < 0.1



from skyfield.api import load

max_days = 7.0         # download again once 7 days old
name = 'stations.csv'  # custom filename, not 'gp.php'

base = 'https://celestrak.org/NORAD/elements/gp.php'
url = base + '?GROUP=stations&FORMAT=csv'

if not load.exists(name) or load.days_old(name) >= max_days:
    load.download(url, filename=name)



# Runtime was around 1 minute 4 seconds under the old find-discrete
# regime; with the new find_risings() and find_settings() routines, this
# script now takes around 5 seconds.  Nice.

from skyfield.api import load, wgs84
from skyfield import almanac

ts = load.timescale()
eph = load('de421.bsp')
grand_canyon_village = wgs84.latlon(+36.0544, -112.1401)
observer = eph['earth'] + grand_canyon_village
tz_hours = -7
start = ts.utc(1986, 1, 17, -tz_hours)
#end = ts.utc(1987, 7, 14, -tz_hours)
end = ts.utc(2022, 7, 14, -tz_hours)
sunrises, _ = almanac.find_risings(observer, eph['Sun'], start, end)
sunsets, _ = almanac.find_settings(observer, eph['Sun'], start, end)
lengths = sunsets - sunrises
print('date,hours_sunlight')
for ti, length in zip(sunrises, lengths):
    print('{},{}'.format(ti.utc_strftime('%Y-%m-%d'), length * 24.0))



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from skyfield import almanac
from skyfield.api import load, wgs84
from skyfield.magnitudelib import planetary_magnitude

MONTH_NAMES = '0 Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec'.split()

# Figure out the times of sunset over our range of dates.

eph = load('de421.bsp')
earth, sun, venus = eph['earth'], eph['sun'], eph['venus']
observer = wgs84.latlon(+40.0, 0.0)

ts = load.timescale()
start, end = ts.utc(2021, 3, 7), ts.utc(2022, 2, 7)

f = almanac.sunrise_sunset(eph, observer)
t, y = almanac.find_discrete(start, end, f)
sunsets = (y == 0)
t = t[sunsets]

# For each moment of sunset, ask Skyfield for the month number, the day
# number, and for Venus’s altitude, azimuth, and magnitude.

year, month, day, hour, minute, second = t.utc
month = month.astype(int)
day = day.astype(int)

apparent = (earth + observer).at(t).observe(venus).apparent()
alt, az, distance = apparent.altaz()
x, y = az.degrees, alt.degrees
m = planetary_magnitude(apparent)

# Convert magnitude to marker size, remembering that smaller magnitude
# numbers mean a brighter Venus (and thus a larger marker).

maxmag = max(m)
minmag = min(m)
size = 40 - 30 * (m - minmag) / (maxmag - minmag)

# Start with a smooth curve tracing Venus's motion.

fig, ax = plt.subplots(figsize=[9, 3])
ax.plot(x, y, c='#fff6', zorder=1)

# Next, put a circle representing Venus on the 1st of the month and on
# every fifth day after that.  (Except for the 30th, which would sit too
# close to the 1st of the following month.)

fives = (day % 5 == 1) & (day < 30)
ax.scatter(x[fives], y[fives], size[fives], 'white',
           edgecolor='black', linewidth=0.25, zorder=2)

# Put day and month labels off to the sides of the curve.

offset_x, offset_y = 10, 8

for i in np.flatnonzero(fives):
    if i == 0:
        continue  # We can’t compute dx/dy with no previous point.

    # Build a unit vector pointing in the direction Venus is traveling.

    day_i = day[i]
    xi = x[i]
    yi = y[i]
    dx = xi - x[i-1]
    dy = yi - y[i-1]
    length = np.sqrt(dx*dx + dy*dy)
    dx /= length
    dy /= length

    # Offset the text at a right angle to the direction of travel.

    side = 'right' if (year[i], month[i]) < (2021, 10) else 'left'
    if side == 'left':
        xytext = - offset_x*dy, offset_y*dx
    else:
        xytext = offset_x*dy, - offset_y*dx

    # Label the dates 1, 11, and 21.

    if day_i in (1, 11, 21):
        ax.annotate(day_i, (xi, yi), c='white', ha='center', va='center',
                    textcoords='offset points', xytext=xytext, size=8)

    # On the 15th of each month, put the month name.

    if day_i == 16:
        name = MONTH_NAMES[month[i]]
        ax.annotate(name, (xi, yi), c='white', ha='center', va='center',
                    textcoords='offset points', xytext=2.2 * np.array(xytext))

# Finally, some decorations.

points = 'N NE E SE S SW W NW'.split()
for i, name in enumerate(points):
    xy = 45 * i, 1
    ax.annotate(name, xy, c='white', ha='center', size=12, weight='bold')

ax.set(
    aspect=1.0,
    title='Venus at sunset for 40°N latitude, April 2021 – January 2022',
    xlabel='Azimuth (°)',
    ylabel='Altitude (°)',
    xlim=(195, 300),
    ylim=(0, max(y) + 10.0),
    xticks=np.arange(210, 300, 15),
)

sky = LinearSegmentedColormap.from_list('sky', ['black', 'blue'])
extent = ax.get_xlim() + ax.get_ylim()
ax.imshow([[0,0], [1,1]], cmap=sky, interpolation='bicubic', extent=extent)

fig.savefig('venus_evening_chart.png')



# -*- coding: utf-8 -*-
"""Routines to solve for circumstances like sunrise, sunset, and moon phase."""

from __future__ import division

import numpy as np
from numpy import cos, sin, zeros_like
from .constants import pi, tau
from .framelib import ecliptic_frame
from .searchlib import find_discrete
from .nutationlib import iau2000b_radians
from .units import Angle

# Not only to support historic code but also for future convenience, let
# folks import the search routine alongside the almanac routines.
find_discrete

# Simple facts.

def phase_angle(ephemeris, body, t):
    """
    .. deprecated:: 1.42
       Use the :meth:`~skyfield.positionlib.ICRF.phase_angle()` position
       method instead.
    """
    p = ephemeris['earth'].at(t).observe(ephemeris[body])
    return p.phase_angle(ephemeris['sun'])

def fraction_illuminated(ephemeris, body, t):
    """
    .. deprecated:: 1.42
       Use the :meth:`~skyfield.positionlib.ICRF.fraction_illuminated()`
       position method instead.
    """
    a = phase_angle(ephemeris, body, t).radians
    return 0.5 * (1.0 + cos(a))

# Discrete circumstances to search.

SEASONS = [
    'Spring',
    'Summer',
    'Autumn',
    'Winter',
]

SEASON_EVENTS = [
    'Vernal Equinox',
    'Summer Solstice',
    'Autumnal Equinox',
    'Winter Solstice',
]

SEASON_EVENTS_NEUTRAL = [
    'March Equinox',
    'June Solstice',
    'September Equinox',
    'December Solstice',
]

def seasons(ephemeris):
    """Build a function of time that returns the quarter of the year.

    The function that this returns will expect a single argument that is
    a :class:`~skyfield.timelib.Time` and will return 0 through 3 for
    the seasons Spring, Summer, Autumn, and Winter.

    """
    earth = ephemeris['earth']
    sun = ephemeris['sun']

    def season_at(t):
        """Return season 0 (Spring) through 3 (Winter) at time `t`."""
        t._nutation_angles_radians = iau2000b_radians(t)
        e = earth.at(t)
        _, slon, _ = e.observe(sun).apparent().frame_latlon(ecliptic_frame)
        return (slon.radians // (tau / 4) % 4).astype(int)

    season_at.step_days = 90.0
    return season_at

MOON_PHASES = [
    'New Moon',
    'First Quarter',
    'Full Moon',
    'Last Quarter',
]

def moon_phase(ephemeris, t):
    """Return the Moon phase 0°–360° at time ``t``, where 180° is Full Moon.

    More precisely: this returns an :class:`~skyfield.units.Angle`
    giving the difference between the geocentric apparent ecliptic
    longitudes of the Moon and Sun, constrained to the interval 0°–360°
    (0–𝜏 radians) where 0° is New Moon and 180° is Full Moon.

    """
    e = ephemeris['earth'].at(t)
    moon, sun = ephemeris['moon'], ephemeris['sun']
    _, mlon, _ = e.observe(moon).apparent().frame_latlon(ecliptic_frame)
    _, slon, _ = e.observe(sun).apparent().frame_latlon(ecliptic_frame)
    return Angle(radians=(mlon.radians - slon.radians) % tau)

def moon_phases(ephemeris):
    """Build a function of time that returns the moon phase 0 through 3.

    The function that this returns will expect a single argument that is
    a :class:`~skyfield.timelib.Time` and will return the phase of the
    moon as an integer.  See the accompanying array ``MOON_PHASES`` if
    you want to give string names to each phase.

    """
    earth = ephemeris['earth']
    moon = ephemeris['moon']
    sun = ephemeris['sun']

    def moon_phase_at(t):
        """Return the phase of the moon 0 through 3 at time `t`."""
        t._nutation_angles_radians = iau2000b_radians(t)
        e = earth.at(t)
        _, mlon, _ = e.observe(moon).apparent().frame_latlon(ecliptic_frame)
        _, slon, _ = e.observe(sun).apparent().frame_latlon(ecliptic_frame)
        return ((mlon.radians - slon.radians) // (tau / 4) % 4).astype(int)

    moon_phase_at.step_days = 7.0  # one lunar phase per week
    return moon_phase_at

MOON_NODES = [
    'descending',
    'ascending',
]

def moon_nodes(ephemeris):
    """Build a function of time that identifies lunar nodes.

    This returns a function taking a :class:`~skyfield.timelib.Time` and
    returning ``True`` if the Moon is above the ecliptic else ``False``.
    See :ref:`lunar-nodes` for how to use this routine.

    """
    earth = ephemeris['earth']
    moon = ephemeris['moon']

    def moon_node_at(t):
        """Return the phase of the moon 0 through 3 at time `t`."""
        e = earth.at(t)
        lat, _, _ = e.observe(moon).apparent().frame_latlon(ecliptic_frame)
        return lat.radians > 0.0

    moon_node_at.step_days = 12.0  # 2000-2050: closest nodes 12.38 days apart
    return moon_node_at

CONJUNCTIONS = [
    'conjunction',
    'opposition',
]

def oppositions_conjunctions(ephemeris, target):
    """Build a function to find oppositions and conjunctions with the Sun.

    See :ref:`oppositions-conjunctions` for how to call this routine and
    interpret the results.

    """
    earth_at = ephemeris['earth'].at
    sun = ephemeris['sun']

    def leading_or_trailing(t):
        """Return whether the target is east or west of the Sun."""
        e = earth_at(t)
        _, slon, _ = e.observe(sun).apparent().frame_latlon(ecliptic_frame)
        _, tlon, _ = e.observe(target).apparent().frame_latlon(ecliptic_frame)
        return ((slon.radians - tlon.radians) / pi % 2.0).astype('int8')

    if target.target == 301:
        leading_or_trailing.step_days = 14  # Moon
    else:
        leading_or_trailing.step_days = 40  # Mercury (the fastest planet)
    return leading_or_trailing

MERIDIAN_TRANSITS = ['Antimeridian transit', 'Meridian transit']

def meridian_transits(ephemeris, target, topos):
    """Build a function of time for finding when a body transits the meridian.

    The returned function accepts a :class:`~skyfield.timelib.Time`
    argument and returns ``True`` if the ``target`` body is west of the
    observer’s meridian at that time, and otherwise returns ``False.``
    See :ref:`transits` for how to use this to search for a body’s
    meridian transits and antimeridian transits.

    """
    topos_at = (ephemeris['earth'] + topos).at

    def west_of_meridian_at(t):
        """Return `True` if the target is west of the observer’s meridian."""
        t._nutation_angles_radians = iau2000b_radians(t)
        # TODO: should we work to avoid computing Topos position twice?
        # We could grab its hidden GCRS vector and do the trig ourselves.
        # Or there might be something clever we can do with the two raw
        # vectors, skipping the cost of computing spherical coordinates.
        ra1, _, _ = topos.at(t).radec(epoch='date')
        ra2, _, _ = topos_at(t).observe(target).apparent().radec(epoch='date')
        return (ra1.radians - ra2.radians) % tau < pi

    west_of_meridian_at.step_days = 0.4  # twice a day
    return west_of_meridian_at

def sunrise_sunset(ephemeris, topos):
    """Build a function of time that returns whether the Sun is up.

    The function that is returned will expect a single argument that is
    a :class:`~skyfield.timelib.Time`, and will return ``True`` if the
    sun is up, else ``False``.

    Skyfield uses the same definition as the United States Naval
    Observatory: the Sun is up when its center is 0.8333 degrees below
    the horizon, which accounts for both its apparent radius of around
    16 arcminutes and also for the 34 arcminutes by which atmospheric
    refraction on average lifts the image of the Sun.

    If you need to provide a custom value for refraction, adjust the
    estimate of the Sun’s radius, or account for a vantage point above
    the Earth’s surface, see :ref:`risings-and-settings` to learn about
    the more versatile :func:`~skyfield.almanac.risings_and_settings()`
    routine.

    """
    sun = ephemeris['sun']
    topos_at = (ephemeris['earth'] + topos).at

    def is_sun_up_at(t):
        """Return `True` if the sun has risen by time `t`.

        The Sun has risen if its altitude above the horizon is greater
        than -0.8333 degrees.

        """
        t._nutation_angles_radians = iau2000b_radians(t)
        return topos_at(t).observe(sun).apparent().altaz()[0].degrees >= -0.8333

    is_sun_up_at.step_days = 0.04  # catch days at least an hour long
    return is_sun_up_at

TWILIGHTS = {
    0: 'Night',
    1: 'Astronomical twilight',
    2: 'Nautical twilight',
    3: 'Civil twilight',
    4: 'Day',
}

def dark_twilight_day(ephemeris, topos):
    """Build a function of time returning whether it is dark, twilight, or day.

    The function that this returns will expect a single argument that is
    a :class:`~skyfield.timelib.Time` and will return:

    | 0 — Dark of night.
    | 1 — Astronomical twilight.
    | 2 — Nautical twilight.
    | 3 — Civil twilight.
    | 4 — Sun is up.

    """
    sun = ephemeris['sun']
    topos_at = (ephemeris['earth'] + topos).at

    def is_it_dark_twilight_day_at(t):
        """Return whether the Sun is up, down, or whether there is twilight."""
        t._nutation_angles_radians = iau2000b_radians(t)
        degrees = topos_at(t).observe(sun).apparent().altaz()[0].degrees
        r = zeros_like(degrees, int)
        r[degrees >= -18.0] = 1
        r[degrees >= -12.0] = 2
        r[degrees >= -6.0] = 3
        r[degrees >= -0.8333] = 4
        return r

    is_it_dark_twilight_day_at.step_days = 0.04  # catch days at least an hour long
    return is_it_dark_twilight_day_at

def risings_and_settings(ephemeris, target, topos,
                         horizon_degrees=-34.0/60.0, radius_degrees=0): #?
    """Build a function of time that returns whether a body is up.

    This returns a function taking a :class:`~skyfield.timelib.Time`
    argument returning ``True`` if the body’s altazimuth altitude angle
    plus ``radius_degrees`` is greater than ``horizon_degrees``, else
    ``False``.  See :ref:`risings-and-settings` to learn about how to
    search for risings and settings, and to see more about using the
    parameters ``horizon_degrees`` and ``radius_degrees``.

    """
    topos_at = (ephemeris['earth'] + topos).at
    h = horizon_degrees - radius_degrees

    def is_body_up_at(t):
        """Return `True` if the target has risen by time `t`."""
        t._nutation_angles_radians = iau2000b_radians(t)
        return topos_at(t).observe(target).apparent().altaz()[0].degrees > h

    is_body_up_at.step_days = 0.25
    return is_body_up_at

# Direct-search routines using geometry, that don't need find_discrete().

def _fastify(t):
    t._nutation_angles_radians = iau2000b_radians(t)

def _setting_hour_angle(latitude, declination, altitude_radians):
    """Return the hour angle, in radians, when a body reaches the horizon.

    Given the latitude of an observer, and the declination of a target,
    return the positive hour angle at which the body will set below the
    horizon, where the horizon is specified as `altitude_radians` above
    (positive) or below (negative) the great circle of zero altitude.

    """
    lat = latitude.radians
    dec = declination.radians
    numerator = sin(altitude_radians) - sin(lat) * sin(dec)
    denominator = cos(lat) * cos(dec)
    ha = np.arccos(np.clip(numerator / denominator, -1.0, 1.0))
    return ha

def _rising_hour_angle(latitude, declination, altitude_radians):
    return - _setting_hour_angle(latitude, declination, altitude_radians)

def _transit_ha(latitude, declination, altitude_radians):
    return 0.0

# Per https://aa.usno.navy.mil/faq/RST_defs we estimate 34 arcminutes of
# atmospheric refraction and 16 arcminutes for the radius of the Sun.
_sun_horizon_radians = -50.0 / 21600.0 * tau
_refraction_radians = -34.0 / 21600.0 * tau
_moon_radius_m = 1.7374e6

def _find(observer, target, start_time, end_time, horizon_degrees, f):
    # Build a function h() that returns the angle above or below the
    # horizon we are aiming for, in radians.
    if horizon_degrees is None:
        tt = getattr(target, 'target', None)
        if tt == 10:
            horizon_radians = _sun_horizon_radians
            h = lambda distance: horizon_radians
        elif tt == 301:
            horizon_radians = _refraction_radians
            h = lambda distance: horizon_radians - _moon_radius_m / distance.m
        else:
            horizon_radians = _refraction_radians
            h = lambda distance: horizon_radians
    else:
        horizon_radians = horizon_degrees / 360.0 * tau
        h = lambda distance: horizon_radians

    geo = observer.vector_functions[-1]  # should we check observer.center?
    latitude = geo.latitude

    # Build an array of times 0.8 days apart, in the hopes that nothing
    # ever rises (or sets or transits) twice within a 0.8-day period.
    ts = start_time.ts
    tt0 = start_time.tt
    tt1 = end_time.tt
    sample_count = int(np.ceil((tt1 - tt0) / 0.8)) + 1
    t = ts.tt_jd(np.linspace(tt0, tt1, sample_count))

    # Determine the target's hour angle and declination at those times.
    _fastify(t)
    ha, dec, distance = observer.at(t).observe(target).apparent().hadec()

    # Invoke our geometry formula: for each time `t`, predict the hour
    # angle at which the target will next reach the horizon, if its
    # declination were to remain constant.
    desired_ha_radians = f(latitude, dec, h(distance))

    # So at each time `t`, how many radians must the sky turn to bring
    # the target to the horizon?
    difference = desired_ha_radians - ha.radians
    difference %= tau

    # We want to return each rising exactly once, so where there are
    # runs of several times `t` that all precede the same rising, let's
    # throw the first few out and keep only the last one.
    i, = np.nonzero(np.diff(difference) > 0.0)

    # When might each rising have actually taken place?  Let's
    # interpolate between the two times that bracket each rising.
    a = difference[i]
    b = tau - difference[i + 1]
    tt = t.tt
    interpolated_tt = (b * tt[i] + a * tt[i+1]) / (a + b)
    t = ts.tt_jd(interpolated_tt)

    ha_per_day = tau            # angle the celestrial sphere rotates in 1 day

    # TODO: How many iterations do we need?  And can we cut down on that
    # number if we use velocity intelligently?  For now, we experiment
    # using the ./design/test_sunrise_moonrise.py script in the
    # repository, that checks both the old Skyfiled routines and this
    # new one against the USNO.  It suggests that 3 iterations is enough
    # for the Moon, the fastest-moving Solar System object, to match.
    for i in 0, 1, 2:
        _fastify(t)
        ha, dec, distance = observer.at(t).observe(target).apparent().hadec()
        desired_ha = f(latitude, dec, h(distance))
        ha_adjustment = desired_ha - ha.radians
        ha_adjustment = (ha_adjustment + pi) % tau - pi
        timebump = ha_adjustment / ha_per_day
        t = ts.tt_jd(t.whole, t.tt_fraction + timebump)

    is_above_horizon = (desired_ha % pi != 0.0)
    return t, is_above_horizon

def find_risings(observer, target, start_time, end_time, horizon_degrees=None):
    """Return the times at which a target rises above the eastern horizon.

    Given an observer on the Earth’s surface, a target like the Sun or
    Moon or a planet, and start and stop :class:`~skyfield.timelib.Time`
    objects, this returns two arrays that have the same length.  The
    first is a :class:`~skyfield.timelib.Time` listing the moments at
    which the target rises.  The second array has ``True`` for each time
    the target really crosses the horizon, and ``False`` when the target
    merely transits without actually touching the horizon.

    See `risings-and-settings` for examples, and `horizon_degrees` for
    how to use the ``horizon_degrees`` argument.

    .. versionadded:: 1.47

    """
    return _find(observer, target, start_time, end_time, horizon_degrees,
                 _rising_hour_angle)

def find_settings(observer, target, start_time, end_time, horizon_degrees=None):
    """Return the times at which a target sets below the western horizon.

    Given an observer on the Earth’s surface, a target like the Sun or
    Moon or a planet, and start and stop :class:`~skyfield.timelib.Time`
    objects, this returns two arrays that have the same length.  The
    first is a :class:`~skyfield.timelib.Time` listing the moments at
    which the target sets.  The second array has ``True`` for each time
    the target really crosses the horizon, and ``False`` when the target
    merely transits without actually touching the horizon.

    See `risings-and-settings` for examples, and `horizon_degrees` for
    how to use the ``horizon_degrees`` argument.

    .. versionadded:: 1.47

    """
    return _find(observer, target, start_time, end_time, horizon_degrees,
                 _setting_hour_angle)

def find_transits(observer, target, start_time, end_time):
    """Return the times at which a target transits across the meridian.

    Given an observer on the Earth’s surface, a target like the Sun or
    Moon or a planet, and start and stop :class:`~skyfield.timelib.Time`
    objects, this returns a :class:`~skyfield.timelib.Time` array
    listing the moments at which the target transits across the
    meridian.

    See `transits` for example code.

    .. versionadded:: 1.47

    """
    t, _ = _find(observer, target, start_time, end_time, 0.0, _transit_ha)
    return t



from skyfield.api import Loader, Topos, Angle, EarthSatellite, Star
from skyfield.timelib import Time
from almanac2 import (meridian_transits, culminations, twilights, 
                       risings_settings, seasons, moon_phases, apsides, nodes,
                       max_ecliptic_latitudes)
from almanac2 import (_ecliptic_lon_diff, _moon_ul_alt, _lha, _alt,
                       _satellite_alt, _ecliptic_lon, _linear_dist, 
                       _ecliptic_lat)
from numpy import ndarray
from functools import partial

# Put your data directory here before running this file:
load = Loader(r'.')

ts = load.timescale()
ephem = load('de430t.bsp')
earth = ephem['earth']
sun = ephem['sun']
moon = ephem['moon']
mars = ephem['mars barycenter']
venus = ephem['venus']

plain_greenwich = Topos(latitude='51.5 N', longitude='0 W')
greenwich = earth + plain_greenwich

iss_tle = """\
1 25544U 98067A   18161.85073725  .00003008  00000-0  52601-4 0  9993
2 25544  51.6418  50.3007 0003338 171.6979 280.7366 15.54163173117534
"""
ISS = EarthSatellite(*iss_tle.splitlines())

sirius = Star(ra_hours=(6, 45, 8.91728), dec_degrees=(-16, 42, 58.0171))

minute = 1/24/60
sec = minute/60
ms = sec/1000


def compare(value, expected_value, epsilon):
    if hasattr(value, '__len__') or hasattr(expected_value, '__len__'):
        assert max(abs(value - expected_value)) <= epsilon
    else:
        assert abs(value - expected_value) <= epsilon


def is_root(f, times, target_f, epsilon):
    left_times = times - epsilon
    right_times = times + epsilon
    left_f = (f(left_times) - target_f + 180)%360 - 180
    right_f = (f(right_times) - target_f + 180)%360 - 180
    
    success_increasing = (left_f <= 0) * (right_f >= 0)
    success_decreasing = (left_f >= 0) * (right_f <= 0)
    success = success_increasing + success_decreasing
    
    if isinstance(success, ndarray):
        return success.all()
    else:
        return success
    
    
def is_extreme(f, times, epsilon):
    left_times = times - epsilon
    right_times = times + epsilon
    center_f = f(times)
    left_f = (f(left_times) - center_f + 180)%360 - 180
    right_f = (f(right_times) - center_f + 180)%360 - 180
    
    is_max = (left_f <= 0) * (right_f <= 0)
    is_min = (left_f >= 0) * (right_f >= 0)
    success = is_max + is_min
    
    if isinstance(success, ndarray):
        return success.all()
    else:
        return success


# Data Source:
# web.archive.org/web/20171223014917/http://aa.usno.navy.mil/data/docs/EarthSeasons.php
def test_seasons():
    t0 = ts.utc(2000)
    t1 = ts.utc(2026)
    times, lons = seasons(earth, t0, t1)
    
    assert ((lons.degrees==0) + (lons.degrees==90) + (lons.degrees==180) + (lons.degrees==270)).all()
    
    assert isinstance(times, Time)
    assert isinstance(lons, Angle)
    assert len(times) == len(lons.radians) == 104
    
    compare(times[0].tt, ts.utc(2000, 3, 20, 7, 35).tt, minute/2)
    compare(times[1].tt, ts.utc(2000, 6, 21, 1, 48).tt, minute/2)
    
    # Check that the found times produce the correct data
    f = partial(_ecliptic_lon, earth, sun)
    assert is_root(f, times.tt, lons.degrees, ms/2)
    

# Data Source:
# http://aa.usno.navy.mil/cgi-bin/aa_phases.pl?year=2017&month=1&day=1&nump=50&format=p
def test_moon_phases():
    t0 = ts.utc(2017)
    t1 = ts.utc(2018)
    times, lon_diffs = moon_phases(moon, t0, t1)
    
    assert ((lon_diffs.degrees==0) + (lon_diffs.degrees==90) + (lon_diffs.degrees==180) + (lon_diffs.degrees==270)).all()

    assert isinstance(times, Time)
    assert isinstance(lon_diffs, Angle)
    assert len(times) == len(lon_diffs.radians) == 49
    
    compare(times[0].tt, ts.utc(2017, 1, 5, 19, 47).tt, minute/2)
    
    # Check that the found times produce the correct data    
    f = partial(_ecliptic_lon_diff, earth, moon, sun)
    assert is_root(f, times.tt, lon_diffs.degrees, ms/2)


class TestRisingsSettings():
    # Data Source:
    # http://aa.usno.navy.mil/cgi-bin/aa_rstablew.pl?ID=AA&year=2017&task=0&place=Greenwich&lon_sign=-1&lon_deg=0&lon_min=0&lat_sign=1&lat_deg=51&lat_min=30&tz=&tz_sign=-1
    def test_sun_risings_settings(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, kinds = risings_settings(greenwich, sun, t0, t1)
        
        assert ((kinds=='rise') + (kinds=='set')).all()
        
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 62
        
        compare(times[0].tt, ts.utc(2017, 1, 1, 8, 6).tt, minute/2)
        
        # Check that the found times produce the correct data    
        f = partial(_alt, greenwich, sun)
        assert is_root(f, times.tt, -50/60, ms/2)
    
    
    # Data Source:   
    # http://aa.usno.navy.mil/cgi-bin/aa_rstablew.pl?ID=AA&year=2017&task=1&place=Greenwich&lon_sign=-1&lon_deg=0&lon_min=0&lat_sign=1&lat_deg=51&lat_min=30&tz=&tz_sign=-1
    def test_moon_risings_settings(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, kinds = risings_settings(greenwich, moon, t0, t1)
        
        assert ((kinds=='rise') + (kinds=='set')).all()
        
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 60
        
        compare(times[0].tt, ts.utc(2017, 1, 1, 9, 46).tt, minute/2)
        
        # Check that the found times produce the correct data
        f = partial(_moon_ul_alt, greenwich, moon)
        assert is_root(f, times.tt, -34/60, ms/2)
        
        
    def test_ISS_risings_settings(self):
        t0 = ts.utc(2017, 6, 1)
        t1 = ts.utc(2017, 6, 2)
        times, kinds = risings_settings(greenwich, ISS, t0, t1)
        
        assert ((kinds=='rise') + (kinds=='set')).all()
        
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 13
        
        # make sure it also works with plain Topos objects
        assert (times==risings_settings(plain_greenwich, ISS, t0, t1)[0]).all()
        
        # Check that the found times produce the correct data
        f = partial(_satellite_alt, plain_greenwich, ISS)
        assert is_root(f, times.tt, -34/60, ms/2)
    
    
    def test_sirius_risings_settings(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, kinds = risings_settings(greenwich, sirius, t0, t1)
        
        assert ((kinds=='rise') + (kinds=='set')).all()
        
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 62
        
        # Check that the found times produce the correct data
        f = partial(_alt, greenwich, sirius)
        assert is_root(f, times.tt, 0, ms/2)
    

class TestTwilights():
    # Data Source:
    # http://aa.usno.navy.mil/cgi-bin/aa_rstablew.pl?ID=AA&year=2017&task=2&place=Greenwich&lon_sign=-1&lon_deg=0&lon_min=0&lat_sign=1&lat_deg=51&lat_min=30&tz=&tz_sign=-1
    def test_civil_twilights(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, am_pm = twilights(greenwich, sun, t0, t1, 'civil')
    
        assert ((am_pm=='am') + (am_pm=='pm')).all()
    
        assert isinstance(times, Time)
        assert isinstance(am_pm, ndarray)
        assert len(times) == len(am_pm) == 62
        
        compare(times[0].tt, ts.utc(2017, 1, 1, 7, 26).tt, minute/2)
        
        # Check that the found times produce the correct data
        f = partial(_alt, greenwich, sun)
        assert is_root(f, times.tt, -6, ms/2)
    
    
    # Data Source:
    # http://aa.usno.navy.mil/cgi-bin/aa_rstablew.pl?ID=AA&year=2017&task=3&place=Greenwich&lon_sign=-1&lon_deg=0&lon_min=0&lat_sign=1&lat_deg=51&lat_min=30&tz=&tz_sign=-1
    def test_nautical_twilights(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, am_pm = twilights(greenwich, sun, t0, t1, 'nautical')
        
        assert ((am_pm=='am') + (am_pm=='pm')).all()
    
        assert isinstance(times, Time)
        assert isinstance(am_pm, ndarray)
        assert len(times) == len(am_pm) == 62
        
        compare(times[0].tt, ts.utc(2017, 1, 1, 6, 43).tt, minute/2)
        
        # Check that the found times produce the correct data
        f = partial(_alt, greenwich, sun)
        assert is_root(f, times.tt, -12, ms/2)
        
        
    # Data Source:
    # http://aa.usno.navy.mil/cgi-bin/aa_rstablew.pl?ID=AA&year=2017&task=4&place=Greenwich&lon_sign=-1&lon_deg=0&lon_min=0&lat_sign=1&lat_deg=51&lat_min=30&tz=&tz_sign=-1
    def test_astronomical_twilights(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, am_pm = twilights(greenwich, sun, t0, t1, 'astronomical')
        
        assert ((am_pm=='am') + (am_pm=='pm')).all()
    
        assert isinstance(times, Time)
        assert isinstance(am_pm, ndarray)
        assert len(times) == len(am_pm) == 62
        
        compare(times[0].tt, ts.utc(2017, 1, 1, 6, 2).tt, minute/2)
        
        # Check that the found times produce the correct data
        f = partial(_alt, greenwich, sun)
        assert is_root(f, times.tt, -18, ms/2)
    

class TestTransits():
    # Data Source:
    # http://aa.usno.navy.mil/cgi-bin/aa_mrst2.pl?form=2&ID=AA&year=2017&month=1&day=1&reps=31&body=10&place=Greenwich&lon_sign=-1&lon_deg=&lon_min=&lon_sec=&lat_sign=1&lat_deg=51&lat_min=30&lat_sec=&height=&tz=&tz_sign=-1
    def test_sun_transits(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, lhas = meridian_transits(greenwich, sun, t0, t1)
        
        assert ((lhas.hours==0) + (lhas.hours==12)).all()
    
        assert isinstance(times, Time)
        assert isinstance(lhas, Angle)
        assert len(times) == len(lhas.radians) == 62
            
        compare(times[1].tt, ts.utc(2017, 1, 1, 12, 4).tt, minute/2)
    
        # Check that the found times produce the correct data
        f = partial(_lha, greenwich, sun)
        assert is_root(f, times.tt, lhas._degrees, ms/2)
        
    
    # Data Source:
    # http://aa.usno.navy.mil/cgi-bin/aa_mrst2.pl?form=2&ID=AA&year=2017&month=1&day=1&reps=31&body=4&place=Greenwich&lon_sign=-1&lon_deg=&lon_min=&lon_sec=&lat_sign=1&lat_deg=51&lat_min=30&lat_sec=&height=&tz=&tz_sign=-1
    def test_mars_transits(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, lhas = meridian_transits(greenwich, mars, t0, t1)
        
        assert ((lhas.hours==0) + (lhas.hours==12)).all()
        
        assert isinstance(times, Time)
        assert isinstance(lhas, Angle)
        assert len(times) == len(lhas.radians) == 62
        
        compare(times[1].tt, ts.utc(2017, 1, 1, 16, 2).tt, minute/2)
        
        # Check that the found times produce the correct data
        f = partial(_lha, greenwich, mars)
        assert is_root(f, times.tt, lhas._degrees, ms/2)
        
    
    def test_sirius_transits(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, lhas = meridian_transits(greenwich, sirius, t0, t1)
        
        assert ((lhas.hours==0) + (lhas.hours==12)).all()
        
        assert isinstance(times, Time)
        assert isinstance(lhas, Angle)
        assert len(times) == len(lhas.radians) == 63
            
        # Check that the found times produce the correct data
        f = partial(_lha, greenwich, sirius)
        assert is_root(f, times.tt, lhas._degrees, ms/2)
    

class TestCulminations():
    def test_sun_culminations(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, kinds = culminations(greenwich, sun, t0, t1)
        
        assert ((kinds=='upper') + (kinds=='lower')).all()
        
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 62
        
        # Check that the found times produce the correct data
        f = partial(_alt, greenwich, sun)        
        assert is_extreme(f, times.tt, 2*ms)
        
        
    def test_moon_culminations(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, kinds = culminations(greenwich, moon, t0, t1)
        
        assert ((kinds=='upper') + (kinds=='lower')).all()
    
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 60
    
        # Check that the found times produce the correct data
        f = partial(_alt, greenwich, moon)    
        assert is_extreme(f, times.tt, 6*ms)
        
        
    def test_ISS_culminations(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 2)
        times, kinds = culminations(greenwich, ISS, t0, t1)
        
        assert ((kinds=='upper') + (kinds=='lower')).all()
    
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 31
        
        # make sure it also works with plain Topos objects
        assert (times==culminations(plain_greenwich, ISS, t0, t1)[0]).all()
    
        # Check that the found times produce the correct data
        f = partial(_satellite_alt, plain_greenwich, ISS)
        assert is_extreme(f, times.tt, 2*ms)
        
    
    def test_sirius_culminations(self):
        t0 = ts.utc(2017, 1, 1)
        t1 = ts.utc(2017, 1, 32)
        times, kinds = culminations(greenwich, sirius, t0, t1)
        
        assert ((kinds=='upper') + (kinds=='lower')).all()
        
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 63
        
        # Check that the found times produce the correct data
        f = partial(_alt, greenwich, sirius)        
        assert is_extreme(f, times.tt, 2*ms)
    
    
class TestApsides():
    # Data Source:
    # web.archive.org/web/20171223014917/http://aa.usno.navy.mil/data/docs/EarthSeasons.php
    def test_earth_apsides(self):
        t0 = ts.utc(2000)
        t1 = ts.utc(2026)
        times, kinds = apsides(earth, sun, t0, t1)
        
        assert ((kinds=='peri') + (kinds=='apo')).all()
        
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 52
    
        compare(times[0].tt, ts.utc(2000, 1, 3, 5, 18).tt, minute/2)
        
        # Check that the found times produce the correct data
        f = partial(_linear_dist, earth, sun)        
        assert is_extreme(f, times.tt, ms)
        
        
    def test_moon_apsides(self):
        t0 = ts.utc(2017)
        t1 = ts.utc(2018)
        times, kinds = apsides(earth, moon, t0, t1)
        
        assert ((kinds=='peri') + (kinds=='apo')).all()
        
        assert isinstance(times, Time)
        assert isinstance(kinds, ndarray)
        assert len(times) == len(kinds) == 26
            
        # Check that the found times produce the correct data
        f = partial(_linear_dist, earth, moon)        
        assert is_extreme(f, times.tt, ms)
    

def test_moon_nodes():
    t0 = ts.utc(2017)
    t1 = ts.utc(2018)
    times, kinds = nodes(moon, earth, t0, t1)
    
    assert ((kinds=='ascending') + (kinds=='descending')).all()
    
    assert isinstance(times, Time)
    assert isinstance(kinds, ndarray)
    assert len(times) == len(kinds) == 27
    
    # Check that the found times produce the correct data
    f = partial(_ecliptic_lat, earth, moon)
    assert is_root(f, times.tt, 0, ms/2)


def test_small_intervals():
    t0 = ts.utc(2018)
    t1 = ts.utc(2018, 1, 1, 1, 1)
    
    times, kinds = meridian_transits(greenwich, moon, t0, t1)
    assert len(times) == len(kinds.radians) == 0

    times, kinds = culminations(greenwich, moon, t0, t1)
    assert len(times) == len(kinds) == 0
    
    times, kinds = risings_settings(greenwich, moon, t0, t1)
    assert len(times) == len(kinds) == 0
    
    times, kinds = twilights(greenwich, sun, t0, t1)
    assert len(times) == len(kinds) == 0
    
    times, kinds = seasons(earth, t0, t1)
    assert len(times) == len(kinds.radians) == 0
    
    times, kinds = moon_phases(moon, t0, t1)
    assert len(times) == len(kinds.radians) == 0

    times, kinds = apsides(moon, earth, t0, t1)
    assert len(times) == len(kinds) == 0
    
    times, kinds = nodes(moon, earth, t0, t1)
    assert len(times) == len(kinds) == 0
    

class TestMaxLatitudes():
    def test_mars_max_ecliptic_latitudes(self):
        t0 = ts.utc(2017)
        t1 = ts.utc(2030)
        times, lats = max_ecliptic_latitudes(mars, sun, t0, t1)
    
        assert isinstance(times, Time)
        assert isinstance(lats, Angle)
        assert len(times) == len(lats.radians) == 14
    
        # Check that the found times produce the correct data
        f = partial(_ecliptic_lat, sun, mars)
        assert is_extreme(f, times.tt, ms)
    
    
    def test_venus_max_ecliptic_latitudes(self):
        t0 = ts.utc(2017)
        t1 = ts.utc(2022)
        times, lats = max_ecliptic_latitudes(venus, sun, t0, t1)
        
        assert isinstance(times, Time)
        assert isinstance(lats, Angle)
        assert len(times) == len(lats.radians) == 16
        
        # Check that the found times produce the correct data
        f = partial(_ecliptic_lat, sun, venus)
        assert is_extreme(f, times.tt, ms)
    
    
    def test_moon_max_ecliptic_latitudes(self):
        t0 = ts.utc(2017, 1)
        t1 = ts.utc(2017, 6)
        times, lats = max_ecliptic_latitudes(earth, moon, t0, t1)
        
        assert isinstance(times, Time)
        assert isinstance(lats, Angle)
        assert len(times) == len(lats.radians) == 11
        
        # Check that the found times produce the correct data
        f = partial(_ecliptic_lat, moon, earth)
        # TODO: Why is this 400ms?
        assert is_extreme(f, times.tt, 400*ms)


if __name__ == '__main__':
    import pytest
    pytest.main(['-s'])



from numpy import array, max
from skyfield.api import load
from skyfield.constants import C_AUDAY
from skyfield.functions import length_of, _reconcile

def main():
    ts = load.timescale()

    cfltt = _correct_for_light_travel_time   # the original
    #cfltt = _correct_for_light_travel_time2   # possible replacement
    fcfltt = _correct_for_light_travel_time4   # possible replacement

    print('==== One time, one observer, one target ====')

    t = ts.utc(2021, 2, 19, 13, 47)
    planets = load('de421.bsp')
    observer = planets['earth'].at(t)
    target = planets['mars']
    r, v, t2, light_time = cfltt(observer, target)
    print(t.shape, observer.position.km.shape, r.shape)

    print('==== N times, one observer, one target ====')

    t = ts.utc(2021, 2, 19, 13, [46, 47, 48, 49])
    planets = load('de421.bsp')
    observer = planets['earth'].at(t)
    target = planets['mars']
    r, v, t2, light_time = cfltt(observer, target)
    print(t.shape, observer.position.km.shape, '->', r.shape)
    print('Here is where the planet wound up:')
    print(r)

    # The above maneuvers work fine even with the old version of the
    # routine.  But to proceed from here, we need to switch.

    print('==== N times, one observers, M targets [TRY: RIGHT SIDE UP] ====')

    t = ts.utc(2021, 2, 19, 13, [46, 47, 48, 49])
    planets = load('de421.bsp')

    earth = planets['earth']
    if 0:
        # Turn Earth into two observation positions.
        earth = planets['earth']
        earth._at = build_multi_at(earth._at)

    observer = earth.at(t)
    print('observer', observer.position.au.shape)

    target = planets['mars']
    target._at = build_multi_at(target._at)  # Turn Mars into 2 planets.
    print('target', target.at(t).position.au.shape)

    # t = ts.tt(t.tt[:,None])  # What if we add a dimension to t?
    # print('t', t.shape)

    r, v, t2, light_time = _correct_for_light_travel_time2(observer, target)
    print(t.shape, observer.position.km.shape, '->', r.shape)

    print('Does it look like a second planet 1 AU away at the same 4 times?')
    print('First planet:')
    print(r[:,:,0])
    print('Second planet:')
    print(r[:,:,1])

    print('==== N times, one observers, M targets [TRY: UPSIDE DOWN] ====')

    t = ts.utc(2021, 2, 19, 13, [46, 47, 48, 49])
    planets = load('de421.bsp')

    earth = planets['earth']
    if 0:
        # Turn Earth into two observation positions.
        earth = planets['earth']
        earth._at = build_multi_at(earth._at)
    observer = earth.at(t)
    print('observer', observer.position.au.shape)

    target = planets['mars']
    target._at = build_multi_at(target._at)  # Turn Mars into 2 planets.
    print('target', target.at(t).position.au.shape)

    # t = ts.tt(t.tt[:,None])  # What if we add a dimension to t?
    # print('t', t.shape)

    r, v, t2, light_time = _correct_for_light_travel_time3(observer, target)
    print(t.shape, observer.position.km.shape, '->', r.shape)

    print('Does it look like a second planet 1 AU away at the same 4 times?')
    print('First planet:')
    print(r[:,:,0])
    print('Second planet:')
    print(r[:,:,1])

    # Okay, so, the whole problem of broadcasting: is the only reason it
    # comes up because we have more targets than observers?  What if we
    # ask folks to expand the observers array dimensions as well, so
    # that NumPy is never faced with arrays of different numbers of
    # dimensions, which is what triggers its unfortunate decision to
    # start matching at the ends of the arrays instead of their
    # beginnings?

    print('==== N times, one observers, M targets'
          ' [TRY: EXTRA OBSERVER DIMENSION] ====')

    # So: time has its usual extra dimension to accommodate 4 times.
    t = ts.utc(2021, 2, 19, 13, [46, 47, 48, 49])
    planets = load('de421.bsp')

    earth = planets['earth']
    observer = earth.at(t)
    print('observer', observer.position.au.shape)
    print('Supplementing dimensions')
    observer.position.au = observer.position.au[:,:,None]
    observer.velocity.au_per_d = observer.velocity.au_per_d[:,:,None]
    print('observer', observer.position.au.shape)

    target = planets['mars']
    target._at = build_multi_at(target._at)  # Turn Mars into 2 planets.
    print('target', target.at(t).position.au.shape)

    # TODO: if we work out how we can expand the time's dimensions, as
    # in the following line, before earth.at(t) without raising an
    # exception, then the observer time would not be lacking a dimension
    # that would need to be expanded in

    t = ts.tt(t.tt[:,None])  # What if we add a dimension to t?
    # print('t !!!!!!!!!!!!!!', t.shape)
    # print('t !!!!!!!!!!!!!!', t.whole.shape)
    # print('t !!!!!!!!!!!!!!', t.tdb_fraction.shape)

    r, v, t2, light_time = _correct_for_light_travel_time4(observer, target)
    print(t.shape, observer.position.au.shape, '->', r.shape)

    print('Does it look like a second planet 1 AU away at the same 4 times?')
    print('First planet:')
    print(r[:,:,0])
    print('Second planet:')
    print(r[:,:,1])

offset = array([0, 1])[None,None,:]  # Dimensions: [xyz, time, offset]

def build_multi_at(_at):
    # Take a single planet position, and pretend that really there are
    # two targets returning their positions (like comets or asteroids).

    def wrapper(t):
        tposition, tvelocity, gcrs_position, message = _at(t)
        print('_at() t.shape', t.shape)
        if len(t.shape) < 2:
            # If the time lacks a second dimension, then let's expand
            # position and velocity by adding a new dimension (...,2) at
            # the bottom, as though there were two planets.
            tposition = tposition[:,:,None] + offset
            tvelocity = tvelocity[:,:,None] + offset
        else:
            # Otherwise, the time's extra dimension will already have us
            # producing what looks like two objects in the bottom
            # dimension!  We can simply apply our position offset.
            assert tposition.shape[-1] == tvelocity.shape[-1] == 2
            tposition = tposition + offset
            tvelocity = tvelocity + offset
        return tposition, tvelocity, gcrs_position, message
    return wrapper

# The original.

def _correct_for_light_travel_time(observer, target):
    """Return a light-time corrected astrometric position and velocity.

    Given an `observer` that is a `Barycentric` position somewhere in
    the solar system, compute where in the sky they will see the body
    `target`, by computing the light-time between them and figuring out
    where `target` was back when the light was leaving it that is now
    reaching the eyes or instruments of the `observer`.

    """
    t = observer.t
    ts = t.ts
    whole = t.whole
    tdb_fraction = t.tdb_fraction

    cposition = observer.position.au
    cvelocity = observer.velocity.au_per_d

    tposition, tvelocity, gcrs_position, message = target._at(t)

    distance = length_of(tposition - cposition)
    light_time0 = 0.0
    for i in range(10):
        light_time = distance / C_AUDAY
        delta = light_time - light_time0
        if abs(max(delta)) < 1e-12:
            break

        # We assume a light travel time of at most a couple of days.  A
        # longer light travel time would best be split into a whole and
        # fraction, for adding to the whole and fraction of TDB.
        t2 = ts.tdb_jd(whole, tdb_fraction - light_time)

        tposition, tvelocity, gcrs_position, message = target._at(t2)
        distance = length_of(tposition - cposition)
        light_time0 = light_time
    else:
        raise ValueError('light-travel time failed to converge')
    return tposition - cposition, tvelocity - cvelocity, t, light_time

# Try allowing vectors while keeping things "right side up" with x,y,z
# at the top dimension.

def sub(a, b):
    return (a.T - b.T).T

def _correct_for_light_travel_time2(observer, target):
    #
    # This version makes subtraction work by replacing normal NumPy
    # broadcasting subtraction with a sub() function of our own that
    # broadcasts in the other direction.
    #
    t = observer.t
    ts = t.ts
    whole = t.whole
    tdb_fraction = t.tdb_fraction

    cposition = observer.position.au
    cvelocity = observer.velocity.au_per_d

    tposition, tvelocity, gcrs_position, message = target._at(t)

    distance = length_of(sub(tposition, cposition))

    print('distance', distance.shape)  # (t, targets)

    light_time0 = 0.0
    for i in range(10):
        light_time = distance / C_AUDAY   # GOOD: scalar
        delta = light_time - light_time0  # GOOD first time: scalar; 2nd: ?
        if abs(max(delta)) < 1e-12:
            break

        # We assume a light travel time of at most a couple of days.  A
        # longer light travel time would best be split into a whole and
        # fraction, for adding to the whole and fraction of TDB.
        print('tdb_fraction', tdb_fraction.shape, '[ORIGINAL]')
        print('light_time', light_time.shape)
        diff = sub(tdb_fraction, light_time)
        print('sub()', diff.shape)  # (4,2)? YES!!!
        print('whole', whole.shape)  # (4,)? YES!!!
        # whole, diff = _reconcile(whole, diff)  # Winds up not needed?
        # print('sub()', diff.shape)  # (4,2)
        # print('whole', whole.shape)  # (4,1)
        t2 = ts.tdb_jd(whole, diff)
        print('t2', t2.shape)  # (4, 1)? Why not (4, 2)? Because it prints top.

        tposition, tvelocity, gcrs_position, message = target._at(t2)
        print('tposition', tposition.shape)  # Needs to be 3,4,2
        distance = length_of(sub(tposition, cposition))
        light_time0 = light_time
        #exit()
    else:
        raise ValueError('light-travel time failed to converge')
    return sub(tposition, cposition), sub(tvelocity, cvelocity), t, light_time

# What if we turn things upside down and then do normal subtraction?

def _correct_for_light_travel_time3(observer, target):
    #
    # This version uses normal NumPy subtraction with its default
    # broadcasting, which it makes work by always turning the position
    # vectors upside down when it receives them from other parts of
    # Skyfield, then turning them back over when its own computations
    # are done.
    #
    t = observer.t
    ts = t.ts
    whole = t.whole
    tdb_fraction = t.tdb_fraction

    cposition = observer.position.au
    cvelocity = observer.velocity.au_per_d

    cposition = cposition.T
    cvelocity = cvelocity.T

    tposition, tvelocity, gcrs_position, message = target._at(t)
    tposition = tposition.T
    tvelocity = tvelocity.T

    distance = length_of((tposition - cposition).T).T
    light_time0 = 0.0
    for i in range(10):
        light_time = distance / C_AUDAY
        delta = light_time - light_time0
        if abs(max(delta)) < 1e-12:
            break

        # We assume a light travel time of at most a couple of days.  A
        # longer light travel time would best be split into a whole and
        # fraction, for adding to the whole and fraction of TDB.
        t2 = ts.tdb_jd(whole, (tdb_fraction - light_time).T)

        tposition, tvelocity, gcrs_position, message = target._at(t2)
        tposition = tposition.T
        tvelocity = tvelocity.T
        distance = length_of((tposition - cposition).T).T
        light_time0 = light_time
    else:
        raise ValueError('light-travel time failed to converge')
    tposition = tposition - cposition
    tvelocity = tvelocity - cvelocity
    return tposition.T, tvelocity.T, t, light_time

# What if the observer had an extra dimension so it matches too?

def _correct_for_light_travel_time4(observer, target):
    """Return a light-time corrected astrometric position and velocity.

    Given an `observer` that is a `Barycentric` position somewhere in
    the solar system, compute where in the sky they will see the body
    `target`, by computing the light-time between them and figuring out
    where `target` was back when the light was leaving it that is now
    reaching the eyes or instruments of the `observer`.

    """
    t = observer.t
    ts = t.ts
    whole = t.whole
    tdb_fraction = t.tdb_fraction

    cposition = observer.position.au
    cvelocity = observer.velocity.au_per_d

    print('======', cposition.shape)
    # ?
    print('tdb_fraction before:', tdb_fraction.shape)
    tdb_fraction = tdb_fraction[:,None]
    print('tdb_fraction after:', tdb_fraction.shape)

    tposition, tvelocity, gcrs_position, message = target._at(t)

    distance = length_of(tposition - cposition)
    light_time0 = 0.0
    for i in range(10):
        print('i =', i)
        light_time = distance / C_AUDAY
        delta = light_time - light_time0
        if abs(max(delta)) < 1e-12:
            break

        # We assume a light travel time of at most a couple of days.  A
        # longer light travel time would best be split into a whole and
        # fraction, for adding to the whole and fraction of TDB.
        t2 = ts.tdb_jd(whole, tdb_fraction - light_time)

        tposition, tvelocity, gcrs_position, message = target._at(t2)
        distance = length_of(tposition - cposition)
        light_time0 = light_time
    else:
        raise ValueError('light-travel time failed to converge')
    return tposition - cposition, tvelocity - cvelocity, t, light_time

_reconcile  # So CI will think we used it, whether use above is commented or not

if __name__ == '__main__':
    main()



"""An interface between JPL ephemerides and Skyfield."""

import os
from collections import defaultdict

from jplephem.exceptions import OutOfRangeError
from jplephem.spk import SPK
from jplephem.names import target_name_pairs

from .constants import AU_KM, DAY_S
from .errors import EphemerisRangeError
from .timelib import compute_calendar_date
from .vectorlib import VectorFunction, VectorSum, _jpl_code_name_dict

_jpl_name_code_dict = dict(
    (name, target) for (target, name) in target_name_pairs
)

class SpiceKernel(object):
    """Ephemeris file in NASA .bsp format.

    A "Spacecraft and Planet Kernel" (SPK) file from NASA provides
    |xyz| coordinates for bodies in the Solar System like the Sun,
    planets, moons, and spacecraft.

    You can download a .bsp file yourself and use this class to open it,
    or use the Skyfield ``load()`` function to automatically download a
    popular ephemeris.  Once loaded, you can print this object to the
    screen to see a report on the segments that it includes:

    >>> planets = load('de421.bsp')
    >>> print(planets)
    SPICE kernel file 'de421.bsp' has 15 segments
      JD 2414864.50 - JD 2471184.50  (1899-07-28 through 2053-10-08)
          0 -> 1    SOLAR SYSTEM BARYCENTER -> MERCURY BARYCENTER
          0 -> 2    SOLAR SYSTEM BARYCENTER -> VENUS BARYCENTER
          0 -> 3    SOLAR SYSTEM BARYCENTER -> EARTH BARYCENTER
          0 -> 4    SOLAR SYSTEM BARYCENTER -> MARS BARYCENTER
          0 -> 5    SOLAR SYSTEM BARYCENTER -> JUPITER BARYCENTER
          0 -> 6    SOLAR SYSTEM BARYCENTER -> SATURN BARYCENTER
          0 -> 7    SOLAR SYSTEM BARYCENTER -> URANUS BARYCENTER
          0 -> 8    SOLAR SYSTEM BARYCENTER -> NEPTUNE BARYCENTER
          0 -> 9    SOLAR SYSTEM BARYCENTER -> PLUTO BARYCENTER
          0 -> 10   SOLAR SYSTEM BARYCENTER -> SUN
          3 -> 301  EARTH BARYCENTER -> MOON
          3 -> 399  EARTH BARYCENTER -> EARTH
          1 -> 199  MERCURY BARYCENTER -> MERCURY
          2 -> 299  VENUS BARYCENTER -> VENUS
          4 -> 499  MARS BARYCENTER -> MARS

    To retrieve the one or more vectors necessary to compute the
    position of a body relative to the Solar System barycenter, look up
    the body by its name or official SPICE identifying integer:

    >>> planets['earth']
    <VectorSum of 2 vectors:
     'de421.bsp' segment 0 SOLAR SYSTEM BARYCENTER -> 3 EARTH BARYCENTER
     'de421.bsp' segment 3 EARTH BARYCENTER -> 399 EARTH>
    >>> planets[499]
    <VectorSum of 2 vectors:
     'de421.bsp' segment 0 SOLAR SYSTEM BARYCENTER -> 4 MARS BARYCENTER
     'de421.bsp' segment 4 MARS BARYCENTER -> 499 MARS>

    The result will be a :class:`~skyfield.vectorlib.VectorFunction`
    instance that you can ask for a position at a given input time.

    """
    def __init__(self, path):
        self.path = path
        self.filename = os.path.basename(path)
        self.spk = SPK.open(path)
        self.segments = [SPICESegment(self, s) for s in self.spk.segments]
        self.codes = set(s.center for s in self.segments).union(
                         s.target for s in self.segments)

    def __repr__(self):
        return '<{0} {1!r}>'.format(type(self).__name__, self.path)

    def __str__(self):
        segments = self.spk.segments
        lines = ['SPICE kernel file {0!r} has {1} segments'
                 .format(self.filename, len(segments))]
        format_date = '{0}-{1:02}-{2:02}'.format
        start = end = None
        for s in segments:
            if start != s.start_jd or end != s.end_jd:
                start, end = s.start_jd, s.end_jd
                starts = format_date(*compute_calendar_date(int(start)))
                ends = format_date(*compute_calendar_date(int(end)))
                lines.append('  JD {0:.2f} - JD {1:.2f}  ({2} through {3})'
                             .format(start, end, starts, ends))
            lines.append(_format_segment(s))
        return '\n'.join(lines)

    def close(self):
        """Close this ephemeris file."""
        self.spk.close()

        # In practice, users are not confident the file is really closed
        # unless the metadata also disappears.
        del self.segments[:]
        self.codes.clear()

    def comments(self):
        """Return the comments string of this kernel.

        The resulting string often contains embedded newlines, and is
        formatted for a human reader.

        >>> print(planets.comments())
        ; de421.bsp LOG FILE
        ;
        ; Created 2008-02-12/11:33:34.00.
        ...
        LEAPSECONDS_FILE    = naif0007.tls
        SPK_FILE            = de421.bsp
        ...

        """
        return self.spk.comments()

    def names(self):
        """Return all target names that are valid with this kernel.

        >>> pprint(planets.names())
        {0: ['SOLAR_SYSTEM_BARYCENTER', 'SSB', 'SOLAR SYSTEM BARYCENTER'],
         1: ['MERCURY_BARYCENTER', 'MERCURY BARYCENTER'],
         2: ['VENUS_BARYCENTER', 'VENUS BARYCENTER'],
         3: ['EARTH_BARYCENTER',
             'EMB',
         ...

        The result is a dictionary with target code keys and name lists
        as values.  The last name in each list is the one that Skyfield
        uses when printing information about a body.

        """
        d = defaultdict(list)
        for code, name in target_name_pairs:
            if code in self.codes:
                d[code].append(name)
        return dict(d)

    def decode(self, name):
        """Translate a target name into its integer code.

        >>> planets.decode('Venus')
        299

        Raises ``ValueError`` if you supply an unknown name, or
        ``KeyError`` if the target is missing from this kernel.  You can
        supply an integer code if you already have one and just want to
        check whether it is present in this kernel.

        """
        if isinstance(name, int):
            code = name
        else:
            name = name.upper()
            code = _jpl_name_code_dict.get(name)
            if code is None:
                raise ValueError('unknown SPICE target {0!r}'.format(name))
        if code not in self.codes:
            targets = ', '.join(_format_code_and_name(c) for c in self.codes)
            raise KeyError('kernel {0!r} is missing {1!r} -'
                           ' the targets it supports are: {2}'
                           .format(self.filename, name, targets))
        return code

    def __getitem__(self, target):
        """Return a vector function for computing the location of `target`."""
        target = self.decode(target)
        segments = self.segments
        segment_dict = dict((segment.target, segment) for segment in segments)
        chain = tuple(_center(target, segment_dict))
        if len(chain) == 1:
            return chain[0]
        chain = chain[::-1]
        center = chain[0].center
        target = chain[-1].target
        return VectorSum(center, target, chain)

    def __contains__(self, name_or_code):
        if isinstance(name_or_code, int):
            code = name_or_code
        else:
            code = _jpl_name_code_dict.get(name_or_code.upper())
        return code in self.codes

class SPICESegment(VectorFunction):

    def __new__(cls, ephemeris, spk_segment):
        if spk_segment.data_type == 2:
            return object.__new__(ChebyshevPosition)
        if spk_segment.data_type == 3:
            return object.__new__(ChebyshevPositionVelocity)
        raise ValueError('SPK data type {0} not yet supported'
                         .format(spk_segment.data_type))

    def __init__(self, ephemeris, spk_segment):
        self.ephemeris = ephemeris
        self.center = spk_segment.center
        self.target = spk_segment.target
        self.spk_segment = spk_segment

    @property
    def vector_name(self):
        return '{0!r} segment'.format(self.ephemeris.path)

    def time_range(self, ts):
        s = self.spk_segment
        return ts.tdb_jd(s.start_jd), ts.tdb_jd(s.end_jd)

class ChebyshevPosition(SPICESegment):
    def _at(self, t):
        segment = self.spk_segment
        try:
            position, velocity = segment.compute_and_differentiate(
                t.whole, t.tdb_fraction)
        except OutOfRangeError as e:
            start_time, end_time = self.time_range(t.ts)
            s = '%04d-%02d-%02d' % start_time.tdb_calendar()[:3]
            t = '%04d-%02d-%02d' % end_time.tdb_calendar()[:3]
            text = 'ephemeris segment only covers dates %s through %s' % (s, t)
            mask = e.out_of_range_times
            segment = self.spk_segment
            e = EphemerisRangeError(text, start_time, end_time, mask, segment)
            e.__cause__ = None  # avoid exception chaining in Python 3
            raise e

        return position / AU_KM, velocity / AU_KM, None, None


class ChebyshevPositionVelocity(SPICESegment):
    def _at(self, t):
        pv = self.spk_segment.compute(t.whole, t.tdb_fraction)
        return pv[:3] / AU_KM, pv[3:] * DAY_S / AU_KM, None, None


def _center(code, segment_dict):
    """Starting with `code`, follow segments from target to center."""
    while code in segment_dict:
        segment = segment_dict[code]
        yield segment
        code = segment.center

def _format_code_and_name(code):
    name = _jpl_code_name_dict.get(code, None)
    if name is None:
        return str(code)
    return '{0} {1}'.format(code, name)

def _format_segment(segment):
    cname = _jpl_code_name_dict.get(segment.center, 'unknown')
    tname = _jpl_code_name_dict.get(segment.target, 'unknown')
    return '    {0:3} -> {1:<3}  {2} -> {3}'.format(
        segment.center, segment.target, cname, tname)



from numpy import array, sqrt
from skyfield.earthlib import AU_M, ERAD, reverse_terra, tau

def test_reverse_terra_with_zero_iterations():
    # With zero iterations, should return "geocentric" rather than
    # "geodetic" (="correct") longitude and latitude.
    lat, lon, elevation_m = reverse_terra(array([1, 0, 1]), 0, iterations=0)
    one_millimeter = 1e-3
    assert abs(lat - tau / 8) < 1e-16
    assert lon == 0.0
    assert abs(elevation_m - (AU_M * sqrt(2) - ERAD)) < one_millimeter



# -*- coding: utf-8 -*-
"""Routines for computing magnitudes.

Planetary routines adapted from:

https://arxiv.org/pdf/1808.01973.pdf

Which links to:

https://sourceforge.net/projects/planetary-magnitudes/

Which has directories with three successive versions of their magnitude
computation, the most recent of which provides the files on which this
Python code is based:

Ap_Mag_V3.f90
Ap_Mag_Output_V3.txt
Ap_Mag_Input_V3.txt

* ``r`` planet’s distance from the Sun.
* ``delta`` from Earth?
* ``ph_ang`` illumination phase angle (degrees)

"""
from numpy import array, clip, exp, log10, nan, sin, where
from .constants import RAD2DEG
from .functions import angle_between, length_of
from .naifcodes import _target_name

# See "design/planet_tilts.py" in the Skyfield repository.
_SATURN_POLE = array([0.08547883, 0.07323576, 0.99364475])
_SATURN_POLE_2D = _SATURN_POLE[:, None]
_URANUS_POLE = array([-0.21199958, -0.94155916, -0.26176809])
_URANUS_POLE_2D = _URANUS_POLE[:, None]

def planetary_magnitude(position):
    """Given the position of a planet, return its visual magnitude.

    >>> from skyfield.api import load
    >>> from skyfield.magnitudelib import planetary_magnitude
    >>> ts = load.timescale()
    >>> t = ts.utc(2020, 7, 31)
    >>> eph = load('de421.bsp')
    >>> astrometric = eph['earth'].at(t).observe(eph['jupiter barycenter'])
    >>> print('%.2f' % planetary_magnitude(astrometric))
    -2.73

    The formulae are from `Mallama and Hilton “Computing Apparent
    Planetary Magnitude for the Astronomical Almanac” (2018)
    <https://arxiv.org/pdf/1808.01973.pdf>`_.  Two of the formulae have
    inherent limits:

    * Saturn’s magnitude is unknown and the function will return ``nan``
      (the floating-point value “Not a Number”) if the “illumination
      phase angle” — the angle of the vertex observer-Saturn-Sun —
      exceeds 6.5°.

    * Neptune’s magnitude is unknown and will return ``nan`` if the
      illumination phase angle exceeds 1.9° and the position's date is
      before the year 2000.

    And one formula is not fully implemented (though contributions are
    welcome!):

    * Skyfield does not compute which features on Mars are facing the
      observer, which can introduce an error of ±0.06 magnitude.

    """
    target = position.target
    function = _FUNCTIONS.get(target)
    if function is None:
        name = _target_name(target)
        raise ValueError('cannot compute the magnitude of target %s' % name)

    # Shamelessly treat the Sun as sitting at the Solar System Barycenter.
    sun_to_observer = position.center_barycentric.xyz.au
    observer_to_planet = position.xyz.au
    sun_to_planet = sun_to_observer + observer_to_planet

    r = length_of(sun_to_planet)
    delta = length_of(observer_to_planet)
    ph_ang = angle_between(sun_to_planet, observer_to_planet) * RAD2DEG

    if function is _saturn_magnitude:
        if len(sun_to_planet.shape) > 1:
            pole = _SATURN_POLE_2D
        else:
            pole = _SATURN_POLE

        a = angle_between(pole, sun_to_planet)
        sun_sub_lat = a * RAD2DEG - 90.0

        a = angle_between(pole, observer_to_planet)
        observer_sub_lat = a * RAD2DEG - 90.0

        return function(r, delta, ph_ang, sun_sub_lat, observer_sub_lat)

    if function is _uranus_magnitude:
        if len(sun_to_planet.shape) > 1:
            pole = _URANUS_POLE_2D
        else:
            pole = _URANUS_POLE

        a = angle_between(pole, sun_to_planet)
        sun_sub_lat = a * RAD2DEG - 90.0

        a = angle_between(pole, observer_to_planet)
        observer_sub_lat = a * RAD2DEG - 90.0

        return function(r, delta, ph_ang, sun_sub_lat, observer_sub_lat)

    if function is _neptune_magnitude:
        year = position.t.J
        return function(r, delta, ph_ang, year)

    return function(r, delta, ph_ang)

def _mercury_magnitude(r, delta, ph_ang):
    distance_mag_factor = 5 * log10(r * delta)
    ph_ang_factor = (
        6.3280e-02 * ph_ang
        - 1.6336e-03 * ph_ang**2
        + 3.3644e-05 * ph_ang**3
        - 3.4265e-07 * ph_ang**4
        + 1.6893e-09 * ph_ang**5
        - 3.0334e-12 * ph_ang**6
    )
    return -0.613 + distance_mag_factor + ph_ang_factor

def _venus_magnitude(r, delta, ph_ang):
    distance_mag_factor = 5 * log10(r * delta)
    condition = ph_ang < 163.7
    a0 = where(condition, 0.0, 236.05828 + 4.384)
    a1 = where(condition, - 1.044E-03, - 2.81914E+00)
    a2 = where(condition, + 3.687E-04, + 8.39034E-03)
    a3 = where(condition, - 2.814E-06, 0.0)
    a4 = where(condition, + 8.938E-09, 0.0)
    ph_ang_factor = a4
    for a in a3, a2, a1, a0:
        ph_ang_factor *= ph_ang
        ph_ang_factor += a
    return -4.384 + distance_mag_factor + ph_ang_factor

def _earth_magnitude(r, delta, ph_ang):
    distance_mag_factor = 5 * log10(r * delta)
    ph_ang_factor = -1.060e-03 * ph_ang + 2.054e-04 * ph_ang**2
    return -3.99 + distance_mag_factor + ph_ang_factor

def _mars_magnitude(r, delta, ph_ang):
    r_mag_factor = 2.5 * log10(r * r)
    delta_mag_factor = 2.5 * log10(delta * delta)
    distance_mag_factor = r_mag_factor + delta_mag_factor

    geocentric_phase_angle_limit = 50.0

    condition = ph_ang <= geocentric_phase_angle_limit
    a = where(condition, 2.267E-02, - 0.02573)
    b = where(condition, - 1.302E-04, 0.0003445)
    ph_ang_factor = a * ph_ang + b * ph_ang**2

    # Compute the effective central meridian longitude
    # eff_CM = ( sub_earth_long + sub_sun_long ) / 2.
    # if ( abs ( sub_earth_long - sub_sun_long ) > 180. ):
    #     Eff_CM = Eff_CM + 180.
    # if ( Eff_CM > 360. ):
    #     Eff_CM = Eff_CM - 360.

    # ! Use Stirling interpolation to determine the magnitude correction
    # call Mars_Stirling ( 'R', eff_CM, mag_corr_rot )

    # Convert the ecliptic longitude to Ls
    # Ls = h_ecl_long + Ls_offset
    # if ( Ls > 360. ) Ls = Ls - 360.
    # if ( Ls <   0. ) Ls = Ls + 360.

    # Use Stirling interpolation to determine the magnitude correction
    # call Mars_Stirling ( 'O', Ls, mag_corr_orb )

    # Until effects from Mars rotation are written up:
    mag_corr_rot = 0.0
    mag_corr_orb = 0.0

    # Add factors to determine the apparent magnitude
    ap_mag = where(ph_ang <= geocentric_phase_angle_limit, -1.601, -0.367)
    ap_mag += distance_mag_factor + ph_ang_factor + mag_corr_rot + mag_corr_orb

    return ap_mag

def _jupiter_magnitude(r, delta, ph_ang):
    distance_mag_factor = 5 * log10(r * delta)
    geocentric_phase_angle_limit = 12.0
    ph_ang_pi = ph_ang / 180.0
    ph_ang_factor = where(
        ph_ang <= geocentric_phase_angle_limit,
        (6.16E-04 * ph_ang - 3.7E-04) * ph_ang,
        -2.5 * log10(
            ((((- 1.876 * ph_ang_pi + 2.809) * ph_ang_pi - 0.062) * ph_ang_pi
              - 0.363) * ph_ang_pi - 1.507) * ph_ang_pi + 1.0
        ),
    )
    ap_mag = where(
        ph_ang <= geocentric_phase_angle_limit,
        -9.395 + distance_mag_factor + ph_ang_factor,
        -9.428 + distance_mag_factor + ph_ang_factor,
    )
    return ap_mag

def _saturn_magnitude(r, delta, ph_ang, sun_sub_lat, earth_sub_lat, rings=True):
    # Note that sun_sub_lat and earth_sub_lat should be saturnicentric
    # latitude, not saturnidetic.

    r_mag_factor = 2.5 * log10(r * r)
    delta_mag_factor = 2.5 * log10(delta * delta)
    distance_mag_factor = r_mag_factor + delta_mag_factor

    # Then take the square root of the product of the saturnicentric
    # latitude of the Sun and that of Earth but set to zero when the
    # signs are opposite
    product = sun_sub_lat * earth_sub_lat
    signs_same = product >= 0.0
    square_root = product ** where(signs_same, 0.5, 0.0)  # avoid sqrt(neg)
    sub_lat_geoc = where(signs_same, square_root, 0.0)

    # Compute the effect of phase angle and inclination

    geocentric_phase_angle_limit = 6.5
    geocentric_inclination_limit = 27.0

    is_within_geocentric_bounds = (
        (ph_ang <= geocentric_phase_angle_limit)
        & (sub_lat_geoc <= geocentric_inclination_limit)
    )

    ap_mag = where(
        is_within_geocentric_bounds,
        where(
            rings,

            # Use equation #10 for globe+rings and geocentric circumstances.
            -8.914 - 1.825 * sin(sub_lat_geoc / RAD2DEG) + 0.026 * ph_ang
            - 0.378 * sin(sub_lat_geoc / RAD2DEG) * exp(-2.25 * ph_ang),

            # Use equation #11 for globe-alone and geocentric circumstances
            -8.95 - 3.7e-4 * ph_ang + 6.16e-4 * ph_ang**2,
        ),
        where(
            (ph_ang > geocentric_phase_angle_limit) & (~rings),

            # Use equation #12 for globe-alone beyond geocentric phase
            # angle limit
            -8.94 + 2.446e-4 * ph_ang + 2.672e-4 * ph_ang**2
            - 1.506e-6 * ph_ang**3 +4.767e-9 * ph_ang**4,

            nan,
        ),
    )

    ap_mag = ap_mag + distance_mag_factor
    return ap_mag

def _uranus_magnitude(r, delta, ph_ang,
                      sun_sub_lat_planetog, earth_sub_lat_planetog):
    distance_mag_factor = 5.0 * log10 (r * delta)
    sub_lat_planetog = (abs(sun_sub_lat_planetog)
                        + abs(earth_sub_lat_planetog)) / 2.0
    sub_lat_factor = -0.00084 * sub_lat_planetog
    geocentric_phase_angle_limit = 3.1
    ap_mag = -7.110 + distance_mag_factor + sub_lat_factor
    ap_mag += where(
        ph_ang > geocentric_phase_angle_limit,
        (1.045e-4 * ph_ang + 6.587e-3) * ph_ang,
        0.0,
    )
    return ap_mag

def _neptune_magnitude(r, delta, ph_ang, year):
    r_mag_factor = 2.5 * log10(r * r)
    delta_mag_factor = 2.5 * log10(delta * delta)
    distance_mag_factor = r_mag_factor + delta_mag_factor

    # Equation 16 compute the magnitude at unit distance as a function of time
    ap_mag = clip(-6.89 - 0.0054 * (year - 1980.0), -7.00, -6.89)
    ap_mag += distance_mag_factor

    geocentric_phase_angle_limit = 1.9

    ap_mag = where(
        ph_ang > geocentric_phase_angle_limit,

        # Add phase angle factor from equation 17
        # Check the year because equation 17 only pertains to t > 2000.0
        where(
            year >= 2000.0,
            ap_mag + 7.944e-3 * ph_ang + 9.617e-5 * ph_ang**2,
            nan,
        ),

        # Otherwise leave the value unchanged.
        ap_mag,
    )

    return ap_mag

_FUNCTIONS = {
    199: _mercury_magnitude,
    299: _venus_magnitude,
    399: _earth_magnitude,
    499: _mars_magnitude,
    599: _jupiter_magnitude,
    699: _saturn_magnitude,
    799: _uranus_magnitude,
    899: _neptune_magnitude,

    # Some planets can be reasonably identified with their barycenter.
    1: _mercury_magnitude,
    2: _venus_magnitude,
    4: _mars_magnitude,
    5: _jupiter_magnitude,
    6: _saturn_magnitude,
    7: _uranus_magnitude,
    8: _neptune_magnitude,
}



#!/usr/bin/env python3
"""
Various routines useful when designing and tuning Skyfield’s ∆T curve.

1. `ftp://ftp.iers.org/products/eop/rapid/standard/`_
2. `ftp://ftp.iers.org/products/eop/rapid/standard/csv/`_

* For the position of an object at the celestial equator,
  1 second of time = 15 arcseconds.
  1 millisecond of time = 15 mas.

"""
import sys
from time import time

import matplotlib.pyplot as plt
import numpy as np
from numpy import array, concatenate
from skyfield import timelib
from skyfield.api import load
from skyfield.curvelib import Splines, build_spline_given_ends
from skyfield.data import iers
from skyfield.data.earth_orientation import parse_S15_table
from skyfield.functions import load_bundled_npy
from skyfield.timelib import Time, build_delta_t as build_new_delta_t

delta_t_parabola_stephenson_morrison_hohenkerk_2020_maybe = Splines(
    [1825.0, 1925.0, 0.0, 31.4, 0.0, -320.0])

class E2(list):
    def __init__(self, value):
        self.value = value
    def __eq__(self, other):
        return np.array_equal(other, self.value)

class E():
    def __getitem__(self, i):
        return E2(i)
E = E()

def _cat(*args):
    return concatenate(args, axis=1)

def main(argv):
    work_on_delta_t_discontinuities()
    # compare_splines_to_finals2000_error_bars()
    # compare_old_and_new_delta_t()
    # big_solution_vs_slopes()
    # try_solving_relation_between_lod_and_delta_t()

def work_on_delta_t_discontinuities():
    with load.open('ci/finals2000A.all') as f:
        utc_mjd, dut1 = iers.parse_dut1_from_finals_all(f)
    delta_t_recent, leap_dates, leap_offsets = (
        iers._build_timescale_arrays(utc_mjd, dut1)
    )

    ts = load.timescale()
    t = ts.J(np.arange(-3000 * 12, 4000 * 12) / 12)
    #t = ts.J(np.arange(1700 * 12, 2000 * 12) / 12)

    fig, ax = plt.subplots(1, 1)
    fig.set(size_inches=(5, 5))
    ax.plot(t.J[:-1], np.diff(ts.delta_t_function(t.tt)))
    ax.grid()
    ax.set(xlabel='Year', ylabel='∆T month-to-month change')
    fig.tight_layout()
    fig.savefig('tmp.png')

def do_plot(ax, t, delta_t_function):
    ax.plot(t.J, delta_t_function(t.tt))
    ax.plot(t.J, delta_t_function._long_term_function(t.J))
    ax.grid()

def compare_splines_to_finals2000_error_bars():
    #url = 'http://astro.ukho.gov.uk/nao/lvm/Table-S15-v18.txt'
    url = 'http://astro.ukho.gov.uk/nao/lvm/Table-S15.2020.txt'
    with load.open(url) as f:
        names, columns = parse_S15_table(f)

    i, start_year, end_year, a0, a1, a2, a3 = columns

    with load.open('ci/finals2000A.all') as f:
        utc_mjd, dut1 = iers.parse_dut1_from_finals_all(f)
    delta_t_recent, leap_dates, leap_offsets = (
        iers._build_timescale_arrays(utc_mjd, dut1)
    )

    print('Size of IERS table:', delta_t_recent.shape)
    print('Number of splines:', i.shape)

    #year = [-720, 400, 700]
    #year = np.arange(-720, 2010)
    #year = np.arange(1800, 2010)
    #year = np.arange(1980, 2010)
    year = np.arange(1980, 2010, 0.1)
    interpolate = Splines(start_year, end_year, a3, a2, a1, a0)
    s15_curve = interpolate(year)

    finals_tt, finals_delta_t = delta_t_recent
    ts = load.timescale()
    t = ts.utc(year)
    tt = t.tt

    interpolate = Splines(
        finals_tt[:-1],
        finals_tt[1:],
        finals_delta_t[1:] - finals_delta_t[:-1],
        finals_delta_t[:-1],
    )

    T0 = time()
    finals_curve = interpolate(tt)
    print(time() - T0, 's for interpolate()')

    T0 = time()
    finals_curve2 = np.interp(tt, finals_tt, finals_delta_t)
    print(time() - T0, 's for interp()')

    assert (finals_curve == finals_curve2).all()

    diff = max(abs(s15_curve - finals_curve))
    print('Max difference (seconds, arcseconds):', diff, diff * 15)

def compare_old_and_new_delta_t():
    with load.open('ci/finals2000A.all') as f:
        utc_mjd, dut1 = iers.parse_dut1_from_finals_all(f)
    delta_t_recent, leap_dates, leap_offsets = (
        iers._build_timescale_arrays(utc_mjd, dut1)
    )
    plot_delta_t_functions(
        build_legacy_delta_t(delta_t_recent),
        build_new_delta_t(delta_t_recent),
    )

def plot_delta_t_functions(*functions):
    """Display one or more ∆T curves across timescales short to long."""
    ts = load.timescale()
    bounds = [(1990, 2025), (1650, 2025), (1000, 2500), (-2000, 5500)]
    fig, axes = plt.subplots(len(bounds), 1)
    fig.set(size_inches=(5, 10))
    for (lower, upper), ax in zip(bounds, axes):
        t = ts.utc(np.linspace(lower, upper, 1e3))
        for i, f in enumerate(functions):
            label = chr(ord('A') + i)
            linestyle = '-' if i % 2 else '--'
            ax.plot(t.J, f(t.tt), label=label, linestyle=linestyle)
        ax.set(ylabel='∆T (seconds)')
        ax.grid()
    axes[0].legend()
    axes[-1].set(xlabel='Year')
    fig.tight_layout()
    fig.savefig('tmp.png')

def build_legacy_delta_t(delta_t_recent):
    bundled = _cat(
        load_bundled_npy('morrison_stephenson_deltat.npy')[:,:22],
        load_bundled_npy('historic_deltat.npy'),
    )
    recent_start_time = delta_t_recent[0,0]
    i = np.searchsorted(bundled[0], recent_start_time)
    century = 36524.0
    start_tt = bundled[0,0] - century
    start_J = Time(None, start_tt).J
    end_tt = delta_t_recent[0,-1] + century
    end_J = Time(None, end_tt).J
    parabola = timelib.delta_t_parabola_morrison_stephenson_2004
    delta_t_table = _cat(
        [[start_tt], [parabola(start_J)]],
        bundled[:,:i],
        delta_t_recent,
        [[end_tt], [parabola(end_J)]],
    )
    return timelib.DeltaT(delta_t_table[0], delta_t_table[1], parabola)

def big_solution_vs_slopes():
    # Q: Do we really need the huge move_spline_endpoints() routine that
    #    SymPy figured out for us?  Or can we rebuild a spline over a
    #    different range using just its endpoints and their slopes?
    #
    # A: Wow, we can just use their slopes!
    #
    p = timelib.delta_t_parabola_stephenson_morrison_hohenkerk_2016
    row = p.table[:,0]
    print('Original row:')
    print(row)

    x = np.arange(1700, 2000)
    x0 = 1790
    x1 = 1800

    row2 = move_spline_endpoints(x0, x1, row)
    print('Move endpoints:')
    print(row2)

    row3 = build_spline_given_ends(
        x0, p(x0), p.derivative(x0), x1, p(x1), p.derivative(x1),
    )
    print('From y and slopes:')
    print(row3)

    p2 = Splines(row3)

    fig, ax = plt.subplots(1, 1)
    ax.plot(x, p(x))
    ax.plot(x, p2(x), linestyle='--')
    fig.savefig('tmp.png')

def move_spline_endpoints(new_left, new_right, table_row):
    old_left, old_right, a3, a2, a1, a0 = table_row

    k0, k1 = old_left, old_right
    j0, j1 = new_left, new_right

    u0 = a1/(-k0 + k1)
    u1 = j0*u0
    u2 = k0**2
    u3 = k1**2
    u4 = 2*k0
    u5 = a2/(-k1*u4 + u2 + u3)
    u6 = k0**3
    u7 = 3*k0
    u8 = 3*u2
    u9 = a3/(k1**3 + k1*u8 - u3*u7 - u6)
    u10 = j0**3*u9
    u11 = u4*u5
    u12 = j0*u11
    u13 = u8*u9
    u14 = j0*u13
    u15 = j0**2
    u16 = u15*u5
    u17 = u15*u9
    u18 = u16 - u17*u7
    u19 = j0*j1
    u20 = 2*u19*u5
    u21 = 3*u10
    u22 = 6*k0
    u23 = u19*u22*u9
    u24 = j1*u17
    u25 = 3*u24
    u26 = j1**2
    u27 = u26*u9
    u28 = 3*j0*u27

    b0 = a0 - k0*u0 + u1 + u10 - u12 + u14 + u18 + u2*u5 - u6*u9
    b1 = j1*u0 - j1*u11 + j1*u13 - u1 + u12 - u14 - 2*u16 + u17*u22 + u20 - u21 - u23 + u25
    b2 = u18 - u20 + u21 + u23 - 6*u24 + u26*u5 - u27*u7 + u28
    b3 = j1**3*u9 - u10 + u25 - u28
    return new_left, new_right, b3, b2, b1, b0

def try_solving_moving_endpoints_of_spline():
    import sympy as sy
    sy.init_printing()

    a0, a1, a2, a3, k0, k1, j0, j1, new_t = sy.symbols(
        'a0, a1, a2, a3, k0, k1, j0, j1, new_t')

    # Q: How much simpler is it if we only move one end?
    #j0 = k0  # A: Wow, much simpler!  30+ ops instead of 80+
    #j1 = k1  # A: GADS, not much simpler at all, still 80+ operations.

    years = new_t * (j1 - j0) + j0
    old_t = (years - k0) / (k1 - k0)
    d = (((a3 * old_t + a2) * old_t) + a1) * old_t + a0

    #d = sy.factor(d)
    #d = sy.expand(d)
    #d = sy.simplify(d)
    d = sy.expand(d)

    d = sy.collect(d, new_t)
    b0 = d.coeff(new_t, 0)
    b1 = d.coeff(new_t, 1)
    b2 = d.coeff(new_t, 2)
    b3 = d.coeff(new_t, 3)

    commons, outputs = sy.cse(
        [b0, b1, b2, b3],
        sy.numbered_symbols('u'),
        #optimizations='basic',
    )
    n = 0
    for symbol, expr in commons:
        n += sy.count_ops(expr)
        print(symbol, '=', expr)
    print()
    for i, expr in enumerate(outputs):
        n += sy.count_ops(expr)
        print('b{} = {}'.format(i, expr))
    print('Total operations: {}'.format(n))

def try_solving_relation_between_lod_and_delta_t():
    # From 2016 paper:
    #
    # "The following parabola is a good fit overall to the values of ΔT
    # discussed in §3 and plotted in figures 9, 11–14 (it lies outside
    # the box in figure 10). ΔT= −320.0+(32.5±0.6)((year−1825)/100)^2 s.
    # (4.1)"
    #
    # "The long-term average rate of change in lod derived from the time
    # derivative of equation (4.1) is +1.78±0.03 ms cy−1"
    #
    lod_rate_ms_per_century = +1.78
    lod_rate_s_per_century = lod_rate_ms_per_century / 1000.0
    #
    # ΔT falls further behind by the "lod" amount every single day.
    # There are this many days in a century:
    #
    lod_affect_on_delta_t_per_century = lod_rate_s_per_century * 36525.0
    #
    # Integrate: coefficient * t -> (coefficient/2) t^2.
    #
    delta_t_square_coefficient = 0.5 * lod_affect_on_delta_t_per_century
    #
    # The result is happily close to 32.5!
    #
    print(delta_t_square_coefficient)  # prints: 32.50725

    # But their new Supplement claims to integrate "lod = +1.72 t" but
    # without ever offering the corresponding parabola.  Thus, let's
    # repeat the above calculation for this new "lod" value:
    #
    lod_rate_ms_per_century = +1.72
    lod_rate_s_per_century = lod_rate_ms_per_century / 1000.0
    lod_affect_on_delta_t_per_century = lod_rate_s_per_century * 36525.0
    delta_t_square_coefficient = 0.5 * lod_affect_on_delta_t_per_century
    print(delta_t_square_coefficient)  # prints: 31.4115

if __name__ == '__main__':
    main(sys.argv[1:])



URL = 'https://cdsarc.u-strasbg.fr/ftp/cats/I/239/tyc_main.dat'

PANDAS_MESSAGE = """Skyfield needs Pandas to load the Tycho2 catalog

To load the Tycho2 star catalog, Skyfield needs the Pandas data
analysis toolkit.  Try installing it using your usual Python package
installer, like "pip install pandas" or "conda install pandas".
"""

_COLUMN_NAMES = (
    'Catalog', 'TYC', 'Proxy', 'RAhms', 'DEdms', 'Vmag', '---',
    'r_Vmag', 'RAdeg', 'DEdeg', 'AstroRef', 'Plx', 'pmRA',
    'pmDE', 'e_RAdeg', 'e_DEdeg', 'e_Plx', 'e_pmRA', 'e_pmDE', 'DE:RA',
    'Plx:RA', 'Plx:DE', 'pmRA:RA', 'pmRA:DE', 'pmRA:Plx', 'pmDE:RA',
    'pmDE:DE', 'pmDE:Plx', 'pmDE:pmRA', 'Nastro', 'F2', 'HIP', 'BTmag',
    'e_BTmag', 'VTmag', 'e_VTmag', 'r_BTmag', 'B-V', 'e_B-V', '---(2)',
    'Q', 'Fs', 'Source', 'Nphoto', 'VTscat', 'VTmax', 'VTmin', 'Var',
    'VarFlag', 'MultFlag', 'morePhoto', 'm_HIP', 'PPM', 'HD', 'BD', 'CoD',
    'CPD', 'Remark',
)

def load_dataframe(fobj):
    """Given an open file for ``tyc_main.dat``, return a parsed dataframe.

    If the file is gzipped, it will be automatically uncompressed.

    """
    try:
        from pandas import read_csv
    except ImportError:
        raise ImportError(PANDAS_MESSAGE)

    fobj.seek(0)
    magic = fobj.read(2)
    compression = 'gzip' if (magic == b'\x1f\x8b') else None
    fobj.seek(0)

    df = read_csv(
        fobj, sep='|', names=_COLUMN_NAMES, compression=compression,
        usecols=['TYC', 'Vmag', 'RAdeg', 'DEdeg', 'Plx', 'pmRA', 'pmDE'],
        na_values=['     ', '       ', '        '],
    )
    df.columns = (
        'tyc', 'magnitude', 'ra_degrees', 'dec_degrees',
        'parallax_mas', 'ra_mas_per_year', 'dec_mas_per_year',
    )
    df = df.assign(
        ra_hours = df['ra_degrees'] / 15.0,
        epoch_year = 1991.25,
    )
    return df.set_index('tyc')



"""Various curves."""

from numpy import arange
from ._compatibility import interp
from .descriptorlib import reify
from .functions import _to_array

class Splines(object):
    def __init__(self, table):
        table = _to_array(table)
        if len(table.shape) < 2:  # Let caller provide a single row.
            table = table.reshape(table.shape + (1,))
        self.table = table
        self.lower = lower = table[0]
        self.upper = upper = table[1]
        self._width = upper - lower
        self._n = arange(len(lower))
        self.coefficients = table[2:]

    def __call__(self, x):
        i = interp(x, self.lower, self._n)
        i = i.astype(int)
        t = (x - self.lower[i]) / self._width[i]
        coefficients = iter(self.coefficients)
        value = next(coefficients)[i]
        for c in coefficients:
            value *= t
            value += c[i]
        return value

    @reify
    def derivative(self):
        columns = [self.table[0], self.table[1]]
        coefficients = self.table[2:-1]
        for i, c in enumerate(coefficients):
            n = len(coefficients) - i
            columns.append(n * c / self._width)
        return Splines(columns)

def build_spline_given_ends(x0, y0, slope0, x1, y1, slope1):
    width = x1 - x0
    slope0 = slope0 * width
    slope1 = slope1 * width
    a0 = y0
    a1 = slope0
    a2 = -2*slope0 - slope1 - 3*y0 + 3*y1
    a3 = slope0 + slope1 + 2*y0 - 2*y1
    return x0, x1, a3, a2, a1, a0



#
# This is a quick investigation of the behavior of the geographic
# elevation formula `R / cos(lat) - aC`, which misbehaves at 90° north,
# as pointed out in GitHub issue #842.  As we 'add nines' to 89.0°, here
# is how the quantities behave:
#
# aC: this is the 'local radius of curvature', which converges quickly
#     to 6399593.625758493 m at 89.99999°, and stays there, rock-solid,
#     all the way to 90°.
#
#  R: The radius of the line in the xy-plane connecting the pole with
#     the end of the distance vector, which (obviously) becomes very
#     very small as we approach the pole, though machine precision seems
#     to prevent its ever becoming zero.
#
# cos(lat): Alas, fairly imprecise, as the only way it can tell exactly
#     how far we are from the pole is to look at the very last digits of
#     angles like 89.9999999, which only have a few bits left beneath
#     them to hold their precision.
#
# R / cos(lat): So this is our problem, the value that is going
#     sideways.  For the Earth's surface at the North Pole, converted to
#     meters (* AU_M), it starts at a value very close to aC, then
#     begins to diverge - with `R / cos(lat) - aC` growing by 10× for
#     each step from 89.99 through 89.999999999 before its growth slows
#     - but by that point the error in elevation is already 40 m.
#
# How can we compute the height without recourse to the cosine of the
# latitude?  Drawing a diagram, we have a triangle with one leg that's
# the axis through the poles; one leg that's R, sticking out from the
# axis into the xy-plane; and a hypotenuse that's the radius of
# curvature plus the height above-ground-level of our target.
#
# R / cos(lat) is one way of computing the hypotenuse, yes.
#
# But we can also just use Pythagoras.  We already know one leg is R.
# The other?  It turns out that it's the quantity passed to arctan2() to
# compute the latitude!  Thus, `z + aC * e2_sin_lat`, which we can give
# a name to and pass as an additional return value.
#
# Some common values:
#
# 6399.6 km - Earth radius of curvature at pole
# 6378 km - Earth radius at equator
# 6357 km - Earth radius at pole
# 6335.4 km - Earth radius of curvature at equator

from numpy import arctan2, cos, sin, sqrt
from skyfield.api import load, wgs84, tau
from skyfield.constants import AU_M, DEG2RAD, RAD2DEG

ts = load.timescale()
t = ts.utc(2023, 3, 2, 4, 13, 0)
self = wgs84  # 'self', as though inside method

for nines in range(16):
    #
    # Code below is adapted from geographic_position_of().
    #
    latitude = 90.0 - 1.0 / 10.0 ** nines
    # latitude = 1.0 - 1.0 / 10.0 ** nines  # To test behavior at equator.

    position = wgs84.latlon(latitude, 0.0, 0.0).at(t)
    xyz_au, x, y, R, aC, hyp, lat = self._compute_latitude(position)
    _, _, z = xyz_au

    # Official formula: jumps around, because R / cos(lat) is jumping
    # around, as both R and cos(lat) are becoming very small.
    height_au = R / cos(lat) - aC
    height_m = height_au * AU_M

    # So, new formula, building the hypotenuse with Pythagoras.
    thing = sqrt(hyp * hyp + R * R)
    height2_au = thing - aC
    height2_m = height2_au * AU_M

    out = lat * RAD2DEG
    print('{:<17s} {:<17s} {:20.12f} {:20.12f}'.format(
        str(latitude),
        str(out),
        height_m,
        height2_m,
    ))



"""Print a table of how accurate IAU 2000B is with fewer coefficients."""

import matplotlib.pyplot as plt
from numpy import sqrt
from skyfield.api import load
from skyfield.nutationlib import iau2000a, iau2000b

ts = load.timescale()
#t = ts.utc(2000, 1, range(366 * 50))
t = ts.utc(1995, 1, range(366 * (2050 - 1995)))
print(t[0].utc_jpl())
print(t[-1].utc_jpl())

dpsi_a, depsilon_a = iau2000a(t.tt)

for i in list(range(0, 10)) + list(range(10, 71, 10)) + [77]:
    dpsi_b, depsilon_b = iau2000b(t.tt, i)
    diff1 = abs(dpsi_b - dpsi_a).max()
    diff2 = abs(depsilon_b - depsilon_a).max()
    diff_mas = sqrt(diff1 * diff1 + diff2 * diff2) * 1e-4
    if diff_mas > 1000:
        print(i, '%.3f' % (diff_mas / 1e3), 'arcseconds')
    elif diff_mas > 100:
        print(i, int(diff_mas + 0.999), 'mas')
    elif diff_mas > 10:
        print(i, '%.1f' % diff_mas, 'mas')
    else:
        print(i, '%.2f' % diff_mas, 'mas')



from skyfield import almanac, api
from skyfield.api import load, tau

# http://slittlefair.staff.shef.ac.uk/teaching/phy115/session3/page4/page6/page6.html

# sin ALT = sin LAT * sin DEC + cos LAT * cos DEC * cos HA
# sin LAT * sin DEC + cos LAT * cos DEC * cos HA = sin ALT
# cos LAT * cos DEC * cos HA = sin ALT - sin LAT * sin DEC
# cos HA = (sin ALT - sin LAT * sin DEC) / (cos LAT * cos DEC)
# HA = arrcos((sin ALT - sin LAT * sin DEC) / (cos LAT * cos DEC))

#cos HA = (sin ALT - sin LAT sin DEC) / (cos LAT cos DEC)

from numpy import arccos, sin, clip, cos

def _sunrise_hour_angle_radians(latitude, declination, altitude_radians):
    """Return the hour angle at which a body will reach a given altitude.

    Given the latitude of an observer, and the declination of a target,
    return the positive hour angle at which the body will set below the
    horizon, where the horizon is specified as `altitude_radians` above
    (positive) or below (negative) the great circle of zero altitude.

    """
    lat = latitude.radians
    dec = declination.radians
    numerator = sin(altitude_radians) - sin(lat) * sin(dec)
    denominator = cos(lat) * cos(dec)
    ha = arccos(clip(numerator / denominator, -1.0, 1.0))
    return ha

import numpy as np

from skyfield.nutationlib import iau2000b_radians
def fastify(t):
    t._nutation_angles_radians = iau2000b_radians(t)

def find_sunrise(observer, target, horizon_degrees, start_time, end_time):
    geo = observer.vector_functions[-1]
    latitude = geo.latitude
    h = horizon_degrees / 360.0 * tau

    # Build an array of times 0.8 days apart from start_time to end_time.
    ts = start_time.ts
    t = ts.tt_jd(np.arange(start_time.tt, end_time.tt, 0.8))

    # Determine the target's hour angle and declination at those times.
    fastify(t)
    ha, dec, _ = observer.at(t).observe(target).apparent().hadec()

    # Invoke our geometry formula: for each time `t`, predict the hour
    # angle at which the target will next reach the horizon, if its
    # declination were to remain constant.
    setting_ha = _sunrise_hour_angle_radians(latitude, dec, h)
    rising_radians = - setting_ha

    # So at each time `t`, how many radians is the target's hour angle
    # from the target's next 'ideal' rising?
    difference = ha.radians - rising_radians
    difference %= tau

    # We want to return each rising exactly once, so where there are
    # runs of several times `t` that all precede the same rising, let's
    # throw the first few out and keep only the last one.
    i, = np.nonzero(np.diff(difference) < 0.0)

    # When might each rising have actually taken place?  Let's
    # interpolate between the two times that bracket each rising.
    a = tau - difference[i]
    b = difference[i + 1]
    tt = t.tt
    interpolated_tt = (b * tt[i] + a * tt[i+1]) / (a + b)
    t = ts.tt_jd(interpolated_tt)

    ha_per_day = tau

    for i in 0, 1:
        fastify(t)
        ha, dec, _ = observer.at(t).observe(target).apparent().hadec()
        desired_ha = - _sunrise_hour_angle_radians(latitude, dec, h)
        ha_adjustment = desired_ha - ha.radians
        timebump = ha_adjustment / ha_per_day
        t = ts.tt_jd(t.whole, t.tt_fraction + timebump)

    is_above_horizon = (desired_ha != 0.0)
    return t, is_above_horizon

#find_sunrise(earth + bluffton, sun, stdalt, t0, t1)

def main():
    ts = load.timescale()
    eph = load('de421.bsp')
    earth = eph['earth']
    bluffton = api.wgs84.latlon(40.8939, -83.8917)
    #bluffton = api.wgs84.latlon(80.8939, -83.8917)

    t0 = ts.utc(2020, 1, 1)
    #t1 = ts.utc(2020, 1, 10)
    t1 = ts.utc(2021, 1, 1)
    # t0 = ts.utc(2018, 9, 12, 4)
    # t1 = ts.utc(2018, 9, 13, 4)

    from time import time

    if 1:
        sun = eph['sun']
        stdalt = -0.8333 #/ 360.0 * tau

        T0 = time()
        f = almanac.sunrise_sunset(eph, bluffton)
        t_old, y = almanac.find_discrete(t0, t1, f)
        DUR_OLD = time() - T0

    else:
        sun = eph['mercury']
        stdalt = -0.8333 #/ 360.0 * tau

        T0 = time()
        f = almanac.risings_and_settings(
            eph, sun, bluffton,
            horizon_degrees=stdalt,
            radius_degrees=0.0,
        )
        t_old, y = almanac.find_discrete(t0, t1, f)
        DUR_OLD = time() - T0

    print('How good are old find_discrete results?')
    alt, az, _ = (earth + bluffton).at(t_old).observe(sun).apparent().altaz()
    diff_degrees = alt.degrees - stdalt
    print(diff_degrees)
    print()

    # print(t.utc_iso())
    # print(y)

    T0 = time()
    t_new, is_above_horizon = find_sunrise(
        earth + bluffton, sun, stdalt, t0, t1,
    )
    print('sum is_above_horizon:', is_above_horizon.sum())
    DUR_NEW = time() - T0

    t_old = t_old[::2]

    print(len(t_old), 'old')
    print(len(t_new), 'new')

    diff_seconds = (t_old.tt - t_new.tt) * 24.0 * 3600.0
    print('Max difference (seconds):', max(
        abs(diff_seconds)
    ))

    alt, az, _ = (earth + bluffton).at(t_new).observe(sun).apparent().altaz()
    diff_degrees = alt.degrees - stdalt

    if 0:
        import matplotlib.pyplot as plt
        fig, (ax, ax2) = plt.subplots(2)
        ax.plot(t_new.J, diff_seconds, label='diff (sec)', linestyle='--')
        ax.grid()
        ax2.plot(t_new.J, diff_degrees, label='altitude °', linestyle='--')
        ax2.grid()
        #ax.set(xlabel='time (s)', ylabel='voltage (mV)', title='Title')
        #ax.set_aspect(aspect=1.0)
        #ax.axhline(1.0)
        #ax.axvline(1.0)
        #plt.legend()
        fig.savefig('tmp.png')

    print(DUR_OLD, DUR_NEW, DUR_OLD / DUR_NEW)

    #print('HA:', HA / tau * 24.0 - 12)
    #print(HA - api.tau)

main()



"""Compute the tilt of Saturn, for planetary_magnitude()."""

from skyfield.api import position_of_radec

# pc = PlanetaryConstants()
# pc.read_text(load('pck00008.tpc'))

# TODO: Figure out how to extract this from "pck00008.tpc".  For now,
# these values are copied and pasted.

BODY699_POLE_RA = 40.589
BODY699_POLE_DEC = 83.537

p = position_of_radec(BODY699_POLE_RA / 15.0, BODY699_POLE_DEC, 1.0).xyz.au
print('Saturn:', p)

BODY799_POLE_RA = 257.311
BODY799_POLE_DEC = -15.175

p = position_of_radec(BODY799_POLE_RA / 15.0, BODY799_POLE_DEC, 1.0).xyz.au
print('Uranus:', p)



"""
On my laptop, this script shows that simply computing the positions and
the sunlit-ness of the ISS for each of the 381 seconds of this pass
takes roughly the same amount of time as mounting a full search for the
moment it passes into shadow.  But the benefit is far greater, because
with almost no additional expense all altitudes and azimuths can also
be computed.

"""
import numpy as np
from skyfield import almanac, api
from skyfield.nutationlib import iau2000b

ts = api.load.timescale()
eph = api.load('de421.bsp')
sat = api.EarthSatellite(
    '1 25544U 98067A   20150.54009403  .00000886  00000-0  23936-4 0  9992',
    '2 25544  51.6447  80.7632 0002782   3.7438 167.9307 15.49399736229124',
)

topos = api.Topos('18.3368 N', '64.7281 W')
t0 = ts.utc(2020, 6, 5)
t1 = ts.utc(2020, 6, 6)

half_second = 0.5 / 24.0 / 3600.0

from time import time

T0 = time()
t, y = sat.find_events(topos, t0, t1, altitude_degrees=10.0)
print(time() - T0, 'seconds to find events:')

for ti, yi in zip(t, y):
    print(' ', ti.utc_jpl(), yi)

offset = 0
assert list(y[offset:offset+3]) == [0, 1, 2]

pass_start, pass_end = t[offset+0], t[offset+2]

def f(t):
    length = t.tt.shape[0]
    t._nutation_angles = iau2000b(t.tt)
    return sat.at(t).is_sunlit(eph)

f.step_days = 0.001

T0 = time()
t2, y2 = almanac.find_discrete(pass_start, pass_end, f, epsilon=half_second)
print(time() - T0, 'seconds to find moment of entering shadow:')

for ti, yi in zip(t2, y2):
    print(' ', ti.utc_jpl(), 'entered sunlight' if yi else 'entered shadow')

one_second = 1.0 / 24.0 / 3600.0
tt = np.arange(pass_start.tt, pass_end.tt, one_second)

T0 = time()
t = ts.tt_jd(tt)
t._nutation_angles = iau2000b(t.tt)
satpos = (sat - topos).at(t)
is_sunlit = satpos.is_sunlit(eph)
DT = time() - T0

print(DT, 'to compute is_sunlit() for every second of {}-second pass'
      .format(len(tt)))

T0 = time()
alt, az, distance = satpos.altaz()
DT = time() - T0

print(DT, 'to re-use the positions to compute altitude and azimuth')

print(__doc__.rstrip())



# Which is faster?

from timeit import timeit

number = 10000
ideas = (
    'where(n<0, nan, sqrt(abs(n)))',
    'where(n<0, nan, sqrt(where(n<0, 0, n)))',
    'n ** where(n<0, nan, 0.5)',
)

print('Scalar:')  # Single where() is faster.

for idea in ideas:
    print(timeit(
        idea,
        number=number,
        setup=(
            'from numpy import nan, sqrt, where;'
            'n = -5.0'
        ),
    ))

print('Big array:')  # Where + abs is fastest.

for idea in ideas:
    print(timeit(
        idea,
        number=number,
        setup=(
            'from numpy import arange, nan, sqrt, where;'
            'n = - arange(5000)'
        ),
    ))



"""Helpers for making Skyfield tests stable."""

import datetime as dt
import sys

from skyfield import earthlib
import skyfield.api
import skyfield.timelib

IS_32_BIT = (sys.maxsize == 0x7fffffff)
_real_datetime_class = dt.datetime

class datetime(dt.datetime):
    _fake_now = None
    @staticmethod
    def now():
        return datetime._fake_now

def setup(utc=(2015, 10, 11, 10)):
    # Importing Pandas raises a TypeError once our fake datetime is
    # installed.  So instead of waiting, let's import it now.
    __import__('pandas')

    fake_now = dt.datetime(*utc)
    skyfield.api.load = skyfield.iokit.Loader('.', verbose=False)
    skyfield.timelib.Timescale._utcnow = lambda self: fake_now
    datetime._fake_now = fake_now
    dt.datetime = datetime

def teardown():
    skyfield.api.load = skyfield.iokit.Loader('.')
    skyfield.timelib.Timescale._utcnow = dt.datetime.utcnow
    dt.datetime = _real_datetime_class

class low_precision_ERA(object):
    """Compute the Earth rotation angle with only a single float for UT1.

    Skyfield now uses two floats ``t.whole`` and ``t.ut1_fraction`` to
    represent the UT1 Julian date, supplying an additional 16 digits of
    precision.  For the Earth rotation angle, which moves very quickly
    per unit time compared to most other astronomical quantities, this
    knocks Skyfield out of agreement with other libraries like NOVAS and
    SOFA that round UT1 to a single float.  Various tests use this
    context manager to make Skyfield match the lower-precision output.

    """
    def __enter__(self):
        self.saved = earthlib.earth_rotation_angle
        earthlib.earth_rotation_angle = self.era
    def __exit__(self, *args):
        earthlib.earth_rotation_angle = self.saved
    def era(self, whole, fraction):
        rounded_single_float = whole + fraction
        return self.saved(rounded_single_float)
